# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods

"""Test cases for the apt_get module"""

from __future__ import unicode_literals
from __future__ import print_function

import logging # pylint: disable=unused-import
import unittest

from helpers import docker, ubuntu

from hpccm.building_blocks.apt_get import apt_get

class Test_apt_get(unittest.TestCase):
    def setUp(self):
        """Disable logging output messages"""
        logging.disable(logging.ERROR)

    @ubuntu
    @docker
    def test_basic(self):
        """Basic apt_get"""
        a = apt_get(ospackages=['gcc', 'g++', 'gfortran'])
        self.assertEqual(str(a),
r'''RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        g++ \
        gcc \
        gfortran && \
    rm -rf /var/lib/apt/lists/*''')

    @ubuntu
    @docker
    def test_add_repo(self):
        """Add repo and key"""
        a = apt_get(keys=['https://www.example.com/key.pub'],
                    ospackages=['example'],
                    repositories=['deb http://www.example.com all main'])
        self.assertEqual(str(a),
r'''RUN wget -qO - https://www.example.com/key.pub | apt-key add - && \
    echo "deb http://www.example.com all main" >> /etc/apt/sources.list.d/hpccm.list && \
    apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        example && \
    rm -rf /var/lib/apt/lists/*''')

    @ubuntu
    @docker
    def test_download(self):
        """Download parameter"""
        a = apt_get(download=True, download_directory='/tmp/download',
                    ospackages=['libibverbs1'])
        self.assertEqual(str(a),
r'''RUN apt-get update -y && \
    mkdir -m 777 -p /tmp/download && cd /tmp/download && \
    DEBIAN_FRONTEND=noninteractive apt-get download -y --no-install-recommends \
        libibverbs1 && \
    rm -rf /var/lib/apt/lists/*''')

    @ubuntu
    @docker
    def test_extract(self):
        """Extract parameter"""
        a = apt_get(download=True, extract='/usr/local/ofed',
                    ospackages=['libibverbs1'])
        self.assertEqual(str(a),
r'''RUN apt-get update -y && \
    mkdir -m 777 -p /var/tmp/apt_get_download && cd /var/tmp/apt_get_download && \
    DEBIAN_FRONTEND=noninteractive apt-get download -y --no-install-recommends \
        libibverbs1 && \
    mkdir -p /usr/local/ofed && \
    find /var/tmp/apt_get_download -regextype posix-extended -type f -regex "/var/tmp/apt_get_download/(libibverbs1).*deb" -exec dpkg --extract {} /usr/local/ofed \; && \
    rm -rf /var/tmp/apt_get_download && \
    rm -rf /var/lib/apt/lists/*''')



# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods

"""Test cases for the annotate module"""

from __future__ import unicode_literals
from __future__ import print_function

import logging # pylint: disable=unused-import
import unittest

from hpccm.templates.annotate import annotate

class Test_annotate(unittest.TestCase):
    def setUp(self):
        """Disable logging output messages"""
        logging.disable(logging.ERROR)

    def test_no_annotations(self):
        """No variables specified"""
        a = annotate()

        self.assertDictEqual(a.annotate_step(), {})

    def test_no_base(self):
        """Basic annotations"""
        a = annotate(annotate=True, base_annotation=False)
        a.add_annotation('A', 'a')
        a.add_annotation('one', 1)

        self.assertDictEqual(a.annotate_step(), {'A': 'a', 'one': '1'})

    def test_default_base(self):
        """Basic annotations"""
        a = annotate(annotate=True, base_annotation=True)
        a.add_annotation('A', 'a')
        a.add_annotation('one', 1)

        self.assertDictEqual(a.annotate_step(), {'hpccm.annotate.A': 'a',
                                                 'hpccm.annotate.one': '1'})

    def test_custom_base(self):
        """Basic annotations"""
        a = annotate(annotate=True, base_annotation='foo')
        a.add_annotation('A', 'a')
        a.add_annotation('one', 1)

        self.assertDictEqual(a.annotate_step(), {'hpccm.foo.A': 'a',
                                                 'hpccm.foo.one': '1'})

    def test_no_annotations(self):
        """Basic annotations"""
        a = annotate(annotate=False)
        a.add_annotation('A', 'a')
        a.add_annotation('one', 1)

        self.assertDictEqual(a.annotate_step(), {})




# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Global configuration
#
# Import this as
# import hpccm.config
#
# And access variables as
# hpccm.config.var

from __future__ import absolute_import

from distutils.version import StrictVersion
import archspec.cpu
import logging
import platform
import sys

from hpccm.common import cpu_arch
from hpccm.common import container_type
from hpccm.common import linux_distro

# Global variables
g_cpu_arch = cpu_arch.X86_64         # CPU architecture
if platform.machine() == 'aarch64':
  g_cpu_arch = cpu_arch.AARCH64
elif platform.machine() == 'ppc64le':
  g_cpu_arch = cpu_arch.PPC64LE
g_cpu_target = None                  # CPU optimization target
g_ctype = container_type.DOCKER      # Container type
g_linux_distro = linux_distro.UBUNTU # Linux distribution
g_linux_version = StrictVersion('16.04') # Linux distribution version
g_singularity_version = StrictVersion('2.6') # Singularity version
g_wd = '/var/tmp' # Working directory

def get_cpu_architecture():
  """Return the architecture string for the currently configured CPU
  architecture, e.g., `aarch64`, `ppc64le`, or `x86_64`.

  """

  this = sys.modules[__name__]
  if this.g_cpu_arch == cpu_arch.AARCH64:
    return 'aarch64'
  elif this.g_cpu_arch == cpu_arch.PPC64LE:
    if this.g_linux_distro == linux_distro.UBUNTU:
      return 'ppc64el'
    else:
      return 'ppc64le'
  elif this.g_cpu_arch == cpu_arch.X86_64:
    return 'x86_64'
  else: # pragma: no cover
    raise RuntimeError('Unrecognized processor architecture')

def get_cpu_optimization_flags(compiler, version='9999'):
  """Return the CPU optimization flags for the target and compiler
  combination.

  # Arguments

  compiler: A compiler family string recognized by archspec.

  version: The version of the compiler.  The default version is
  `9999`, i.e., assume the compiler supports the latest optimization
  flags.
  """
  this = sys.modules[__name__]

  if not this.g_cpu_target:
    return None

  if this.g_cpu_target not in archspec.cpu.TARGETS:
    logging.warning('unrecognized CPU target "{}"'.format(this.g_cpu_target))
    return None

  try:
    return archspec.cpu.TARGETS[this.g_cpu_target].optimization_flags(compiler, version)
  except Exception as e:
    logging.warning('get_cpu_optimization_flags: {}'.format(e))

  return None

def get_format():
  """Return the container format string for the currently configured
  format, e.g., `bash`, `docker`, or `singularity`."""

  this = sys.modules[__name__]

  if this.g_ctype == container_type.BASH:
    return 'bash'
  elif this.g_ctype == container_type.DOCKER:
    return 'docker'
  elif this.g_ctype == container_type.SINGULARITY:
    return 'singularity'
  else: # pragma: no cover
    raise RuntimeError('Unrecognized format')

def set_container_format(ctype):
  """Set the container format

  # Arguments

  ctype (string): 'docker' to specify the Dockerfile format, or
  'singularity' to specify the Singularity definition file format

  # Raises

  RuntimeError: invalid container type argument
  """

  this = sys.modules[__name__]
  if ctype == 'docker':
    this.g_ctype = container_type.DOCKER
  elif ctype == 'singularity':
    this.g_ctype = container_type.SINGULARITY
  else:
    raise RuntimeError('Unrecognized container format: {}'.format(ctype))

def set_cpu_architecture(arch):
  """Set the CPU architecture

  In most cases, the `baseimage` primitive should be relied upon to
  set the CPU architecture.  Only use this function if you really
  know what you are doing.

  # Arguments

  arch (string): Value values are `aarch64`, `ppc64le`, and `x86_64`.
  `arm` and `arm64v8` are aliases for `aarch64`, `power` is an alias
  for `ppc64le`, and `amd64` and `x86` are aliases for `x86_64`.
  """
  this = sys.modules[__name__]
  if arch == 'aarch64' or arch == 'arm' or arch == 'arm64v8':
    this.g_cpu_arch = cpu_arch.AARCH64
  elif arch == 'ppc64le' or arch == 'power':
    this.g_cpu_arch = cpu_arch.PPC64LE
  elif arch == 'x86_64' or arch == 'amd64' or arch == 'x86':
    this.g_cpu_arch = cpu_arch.X86_64
  else:
    logging.warning('Unable to determine the CPU architecture, defaulting to x86_64')
    this.g_cpu_arch = cpu_arch.X86_64

def set_cpu_target(target):
  """Set the CPU optimization target

  # Arguments

  target (string): A CPU microarchitecture string recognized by
  archspec.
  """
  this = sys.modules[__name__]
  this.g_cpu_target = target

def set_linux_distro(distro):

  """Set the Linux distribution and version

  In most cases, the `baseimage` primitive should be relied upon to
  set the Linux distribution.  Only use this function if you really
  know what you are doing.

  # Arguments

  distro (string): Valid values are `centos7`, `centos8`, `rhel7`,
  `rhel8`, `rockylinux8`, `ubuntu16`, `ubuntu18`, and `ubuntu20`.
  `ubuntu` is an alias for `ubuntu16`, `centos` is an alias for
  `centos7`, and `rhel` is an alias for `rhel7`.

  """
  this = sys.modules[__name__]
  if distro == 'centos':
    this.g_linux_distro = linux_distro.CENTOS
    this.g_linux_version = StrictVersion('7.0')
  elif distro == 'centos7':
    this.g_linux_distro = linux_distro.CENTOS
    this.g_linux_version = StrictVersion('7.0')
  elif distro == 'centos8':
    this.g_linux_distro = linux_distro.CENTOS
    this.g_linux_version = StrictVersion('8.0')
  elif distro == 'rhel':
    this.g_linux_distro = linux_distro.RHEL
    this.g_linux_version = StrictVersion('7.0')
  elif distro == 'rhel7':
    this.g_linux_distro = linux_distro.RHEL
    this.g_linux_version = StrictVersion('7.0')
  elif distro == 'rhel8':
    this.g_linux_distro = linux_distro.RHEL
    this.g_linux_version = StrictVersion('8.0')
  elif distro == 'rockylinux8':
    this.g_linux_distro = linux_distro.ROCKYLINUX
    this.g_linux_version = StrictVersion('8.0')
  elif distro == 'ubuntu':
    this.g_linux_distro = linux_distro.UBUNTU
    this.g_linux_version = StrictVersion('16.04')
  elif distro == 'ubuntu16':
    this.g_linux_distro = linux_distro.UBUNTU
    this.g_linux_version = StrictVersion('16.04')
  elif distro == 'ubuntu18':
    this.g_linux_distro = linux_distro.UBUNTU
    this.g_linux_version = StrictVersion('18.04')
  elif distro == 'ubuntu20':
    this.g_linux_distro = linux_distro.UBUNTU
    this.g_linux_version = StrictVersion('20.04')
  else:
    logging.warning('Unable to determine the Linux distribution, defaulting to Ubuntu')
    this.g_linux_distro = linux_distro.UBUNTU
    this.g_linux_version = StrictVersion('16.04')

def set_singularity_version(ver):
  """Set the Singularity definition file format version

  The Singularity definition file format was extended in version 3.2
  to enable multi-stage builds.  However, these changes are not
  backwards compatible.

  # Arguments

  ver (string): Singularity definition file format version.

  """
  this = sys.modules[__name__]
  this.g_singularity_version = StrictVersion(ver)

def set_working_directory(wd):
  """Set the working directory to use for staging inside the container

  # Arguments

  wd (string): working directory path

  """
  this = sys.modules[__name__]
  this.g_wd = wd

def test_cpu_feature_flag(flag):
  """Return True or False depending on whether the CPU supports the
  given feature flag

  # Arguments

  flag: A CPU feature flag, e.g., `avx`.
  """
  this = sys.modules[__name__]

  if this.g_cpu_target not in archspec.cpu.TARGETS:
    logging.warning('unrecognized CPU target "{}"'.format(this.g_cpu_target))
    return False

  if this.g_cpu_target:
    try:
      return flag in archspec.cpu.TARGETS[this.g_cpu_target]
    except Exception as e:
      logging.warning('get_cpu_optimization_flags: {}'.format(e))

  return False



"""
MILC 7.8.1

Contents:
  Ubuntu 16.04
  CUDA version 10.1
  GNU compilers (upstream)
  OFED (upstream)
  OpenMPI version 3.1.4
  QUDA version 0.8.0
"""
# pylint: disable=invalid-name, undefined-variable, used-before-assignment
# pylama: ignore=E0602

gpu_arch = USERARG.get('GPU_ARCH', 'sm_60')

# add docstring to Dockerfile
Stage0 += comment(__doc__.strip(), reformat=False)

###############################################################################
# Devel stage
###############################################################################
Stage0 += baseimage(image='nvcr.io/nvidia/cuda:10.1-devel-ubuntu18.04',
                    _as='devel')

Stage0 += gnu()
Stage0 += cmake(eula=True)
Stage0 += ofed()
Stage0 += openmpi(version='3.1.4')

# build QUDA
Stage0 += packages(ospackages=['ca-certificates', 'git'])
Stage0 += generic_cmake(branch='develop',
                        cmake_opts=['-D CMAKE_BUILD_TYPE=RELEASE',
                                    '-D QUDA_DIRAC_CLOVER=ON',
                                    '-D QUDA_DIRAC_DOMAIN_WALL=ON',
                                    '-D QUDA_DIRAC_STAGGERED=ON',
                                    '-D QUDA_DIRAC_TWISTED_CLOVER=ON',
                                    '-D QUDA_DIRAC_TWISTED_MASS=ON',
                                    '-D QUDA_DIRAC_WILSON=ON',
                                    '-D QUDA_FORCE_GAUGE=ON',
                                    '-D QUDA_FORCE_HISQ=ON',
                                    '-D QUDA_GPU_ARCH={}'.format(gpu_arch),
                                    '-D QUDA_INTERFACE_MILC=ON',
                                    '-D QUDA_INTERFACE_QDP=ON',
                                    '-D QUDA_LINK_HISQ=ON',
                                    '-D QUDA_MPI=ON'],
                        install=False,
                        postinstall=['cp -a /var/tmp/quda/build/* /usr/local/quda'],
                        preconfigure=['mkdir -p /usr/local/quda'],
                        prefix='/usr/local/quda',
                        repository='https://github.com/lattice/quda.git')

# build MILC
Stage0 += generic_build(branch='develop',
                        build=['cp Makefile ks_imp_rhmc',
                               'cd ks_imp_rhmc',
                               'make -j 1 su3_rhmd_hisq \
                                CC=/usr/local/openmpi/bin/mpicc \
                                LD=/usr/local/openmpi/bin/mpicxx \
                                QUDA_HOME=/usr/local/quda \
                                WANTQUDA=true \
                                WANT_GPU=true \
                                WANT_CL_BCG_GPU=true \
                                WANT_FN_CG_GPU=true \
                                WANT_FL_GPU=true \
                                WANT_FF_GPU=true \
                                WANT_GF_GPU=true \
                                MPP=true \
                                PRECISION=2 \
                                WANTQIO=""'],
                        install=['mkdir -p /usr/local/milc/bin',
                                 'cp /var/tmp/milc_qcd/ks_imp_rhmc/su3_rhmd_hisq /usr/local/milc/bin'],
                        prefix='/usr/local/milc',
                        repository='https://github.com/milc-qcd/milc_qcd')
Stage0 += environment(variables={'PATH': '/usr/local/milc/bin:$PATH'})

###############################################################################
# Release stage
###############################################################################
Stage1 += baseimage(image='nvcr.io/nvidia/cuda:10.1-base-ubuntu18.04')

Stage1 += Stage0.runtime(exclude=['generic_cmake'])

Stage1 += environment(variables={'PATH': '/usr/local/milc/bin:$PATH'})



# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods

"""SCI-F building block"""

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

import logging # pylint: disable=unused-import
import os
import posixpath

import hpccm.base_object
import hpccm.config

from hpccm.common import container_type
from hpccm.primitives.comment import comment
from hpccm.primitives.copy import copy
from hpccm.primitives.shell import shell

class scif(hpccm.base_object):
    """The `scif` building blocks installs components using the
    [Scientific Filesystem (SCI-F)](https://sci-f.github.io).

    Other building blocks and / or primitives should be added to
    the `scif` building block using the `+=` syntax.

    If not generating a Singularity definition file, SCI-F should be
    installed using the [`pip`](#pip) building block prior to this
    building block.

    If not generating a Singularity definition file, this module
    creates SCI-F recipe files in the current directory (see also the
    `file` parameter).

    # Parameters

    _arguments: Specify additional [Dockerfile RUN arguments](https://github.com/moby/buildkit/blob/master/frontend/dockerfile/docs/experimental.md) (Docker specific).

    _env: Boolean flag to specify whether the general container
    environment should be also be loaded when executing a SCI-F
    `%appinstall` block.  The default is False (Singularity specific).

    file: The SCI-F recipe file name.  The default value is the name
    parameter with the `.scif` suffix.

    name: The name to use to label the SCI-F application.  This
    parameter is required.

    _native: Boolean flag to specify whether to use the native
    Singularity support for SCI-F when generating Singularity
    definition files.  The default is True (Singularity specific).

    # Examples

    ```python
    pip(packages=['scif'])
    s = scif(name='example')
    s += openmpi(prefix='/scif/apps/example')
    s += shell(commands=[...])
    ```

    """

    __runtime_called = False

    def __init__(self, **kwargs):
        """Initialize scif building block"""

        super(scif, self).__init__(**kwargs)

        self.__appenv = []
        self.__appfiles = []
        self.__apphelp = []
        self.__appinstall = []
        self.__applabels = []
        self.__apprun = []
        self.__apptest = []
        self.__arguments = kwargs.get('_arguments')
        self.__env = kwargs.get('_env', False)
        self.__name = kwargs.get('name', None)
        if not self.__name:
            raise RuntimeError('"name" must be defined')
        self.__native = kwargs.get('_native', True)
        self.__scif_file = kwargs.get('file', '{}.scif'.format(self.__name))

    def __iadd__(self, item):
        """Add the item to the corresponding type list.  Allows "+=" syntax."""
        if isinstance(item, list):
            for i in item:
                self.__add(i)
        else:
            self.__add(item)
        return self

    def __add(self, item):
        """Break the item down into its constituent primitives and append each
        primitive to the appropriate list"""

        primitives = self.__primitives(item)
        for p in primitives:
            ptype = p.__class__.__name__
            if ptype == 'comment':
                self.__apphelp.append(p)
            elif ptype == 'copy':
                self.__appfiles.append(p)
            elif ptype == 'environment':
                self.__appenv.append(p)
            elif ptype == 'label':
                self.__applabels.append(p)
            elif ptype == 'runscript':
                self.__apprun.append(p)
            elif ptype == 'shell':
                if p._test:
                    self.__apptest.append(p)
                else:
                    self.__appinstall.append(p)
            else:
                raise RuntimeError('unrecognized primitive type: {}'.format(ptype))

    def __scif_recipe(self):
        """Generate the SCI-F recipe instructions, merging primitives of the
        same type because SCI-F does not support duplicate
        sections."""

        recipe = []

        if self.__appenv:
            appenv = self.__appenv[0].merge(self.__appenv,
                                            _app=self.__name)
            recipe.append(appenv)

        if self.__appfiles:
            appfiles = self.__appfiles[0].merge(self.__appfiles,
                                                _app=self.__name)
            recipe.append(appfiles)

        if self.__apphelp:
            apphelp = self.__apphelp[0].merge(self.__apphelp,
                                              _app=self.__name)
            recipe.append(apphelp)

        if self.__appinstall:
            appinstall = self.__appinstall[0].merge(self.__appinstall,
                                                    _app=self.__name,
                                                    _appenv=self.__env)
            recipe.append(appinstall)

        if self.__applabels:
            applabels = self.__applabels[0].merge(self.__applabels,
                                                  _app=self.__name)
            recipe.append(applabels)

        if self.__apprun:
            apprun = self.__apprun[0].merge(self.__apprun,
                                            _app=self.__name)
            recipe.append(apprun)

        if self.__apptest:
            apptest = self.__apptest[0].merge(self.__apptest,
                                              _app=self.__name, _test=True)
            recipe.append(apptest)

        return recipe

    def __primitives(self, item):
        """Item is a building block or a primitive.  A building block consists
        of one or more other building blocks or primitives.
        Ultimately, every building block consists of primitives.

        "Flatten" the item to a list of its constituent primitives.

        """
        return [i for i in self.__iter_flatten(item)]

    def __iter_flatten(self, iterable):
        """Recursively flatten"""
        try:
            for i in iter(iterable):
                for f in self.__iter_flatten(i):
                    yield f
        except TypeError:
            # not iterable
            yield iterable

    def __str__(self):
        """String representation of the building block"""

        scif_recipe = self.__scif_recipe()

        if (self.__native and
            hpccm.config.g_ctype == container_type.SINGULARITY):
            # Take advantage of Singularity's native support for SCI-F.
            return '\n'.join(str(x) for x in scif_recipe)
        else:
            # Generate an external SCI-F recipe file and manually call scif

            # Temporarily switch container format to Singularity to write
            # the SCI-F recipe file
            preserved_ctype = hpccm.config.g_ctype
            hpccm.config.set_container_format('singularity')

            logging.info('Writing {}'.format(self.__scif_file))
            with open(self.__scif_file, 'w') as f:
                f.write('\n\n'.join(str(x) for x in scif_recipe))

            # Restore original container format
            hpccm.config.g_ctype = preserved_ctype

            # Container instructions to copy the SCI-F recipe file
            # into the container and then run scif
            c_scif_file = posixpath.join('/scif/recipes',
                                         os.path.basename(self.__scif_file))
            instructions = []
            instructions.append(comment('SCI-F "{}"'.format(self.__name)))
            instructions.append(
                copy(src=self.__scif_file, dest=c_scif_file))
            instructions.append(
                shell(_arguments = self.__arguments,
                      chdir=False,
                      commands=['scif install {}'.format(c_scif_file)]))

            return '\n'.join(str(x) for x in instructions)

    def runtime(self, _from='0'):
        """Generate the set of instructions to install the runtime specific
        components from a build in a previous stage.

        The entire `/scif` directory is copied into the runtime stage
        on the first call.  Subsequent calls do nothing.

        # Examples
        ```python
        s = scif(...)
        Stage0 += s
        Stage1 += s.runtime()
        ```

        """

        if not scif.__runtime_called:
            scif.__runtime_called = True
            return str(copy(_from=_from, src='/scif', dest='/scif'))
        else:
            return ''



# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods

"""Copy primitive"""

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

from distutils.version import StrictVersion
import logging  # pylint: disable=unused-import
import posixpath

import hpccm.config

from hpccm.common import container_type


class copy(object):
    """The `copy` primitive copies files from the host to the container
    image.

    # Parameters

    _app: String containing the [SCI-F](https://www.sylabs.io/guides/2.6/user-guide/reproducible_scif_apps.html)
    identifier.  This also causes the Singularity block to named
    `%appfiles` rather than `%files` (Singularity specific).

    _chown: Set the ownership of the file(s) in the container image
    (Docker specific).

    dest: Path in the container image to copy the file(s)

    files: A dictionary of file pairs, source and destination, to copy
    into the container image.  If specified, has precedence over
    `dest` and `src`.

    _from: Set the source location to a previous build stage rather
    than the host filesystem (Docker specific).

    _mkdir: Boolean flag specifying that the destination directory
    should be created in a separate `%setup` step.  This can be used
    to work around the Singularity limitation that the destination
    directory must exist in the container image prior to copying files
    into the image.  The default is False (Singularity specific).

    _post: Boolean flag specifying that file(s) should be first copied
    to `/` and then moved to the final destination by a `%post` step.
    This can be used to work around the Singularity limitation that
    the destination must exist in the container image prior to copying
    files into the image.  The default is False (Singularity
    specific).

    src: A file, or a list of files, to copy

    # Examples

    ```python
    copy(src='component', dest='/opt/component')
    ```

    ```python
    copy(src=['a', 'b', 'c'], dest='/tmp')
    ```

    ```python
    copy(files={'a': '/tmp/a', 'b': '/opt/b'})
    ```

    """

    def __init__(self, **kwargs):
        """Initialize primitive"""

        #super(copy, self).__init__()

        self._app = kwargs.get('_app', '')  # Singularity specific
        self.__chown = kwargs.get('_chown', '')  # Docker specific
        self.__dest = kwargs.get('dest', '')
        self.__files = kwargs.get('files', {})
        self.__from = kwargs.get('_from', '')  # Docker specific
        self._mkdir = kwargs.get('_mkdir', '')  # Singularity specific
        self._post = kwargs.get('_post', '')  # Singularity specific
        self.__src = kwargs.get('src', '')

        if self._mkdir and self._post:
            logging.error('_mkdir and _post are mutually exclusive!')
            self._post = False # prefer _mkdir

        if self._app and (self._mkdir or self._post):
            logging.error('_app cannot be used with _mkdir or _post!')
            self._mkdir = False # prefer _app
            self._post = False

    def __str__(self):
        """String representation of the primitive"""

        # Build a list of files to make the logic a bit simpler below.
        # The items in the files list are dictionaries with keys 'src'
        # and 'dest'.
        files = []
        if self.__files:
            # Sort to make it deterministic
            files.extend([{'dest': dest, 'src': src}
                          for src, dest in sorted(self.__files.items())])
        elif self.__dest and self.__src:
            files.append({'dest': self.__dest, 'src': self.__src})
        else:
            # No files!
            return ''

        if hpccm.config.g_ctype == container_type.DOCKER:
            if self._app:
                logging.warning('The Singularity specific SCI-F syntax '
                                'was requested. Docker does not have an '
                                'equivalent: using regular COPY!')

            # Format:
            # COPY src1 \
            #     src2 \
            #     src3 \
            #     dest/
            # COPY src1 dest1
            # COPY src2 dest2
            # COPY src3 dest3
            base_inst = 'COPY '

            if self.__chown:
                base_inst = base_inst + '--chown={} '.format(self.__chown)

            if self.__from:
                base_inst = base_inst + '--from={} '.format(self.__from)

            # Docker does not have the notion of copying a set of
            # files to different locations inside the container in a
            # single instruction.  So generate multiple COPY
            # instructions in that case.
            instructions = []
            for pair in files:
                dest = pair['dest']
                src = pair['src']
                c = [base_inst]

                if isinstance(src, list):
                    c[0] = c[0] + src[0]
                    c.extend(['    {}'.format(x) for x in src[1:]])
                    # Docker requires a trailing slash.  Add one if missing.
                    c.append('    {}'.format(posixpath.join(dest, '')))
                else:
                    c[0] = c[0] + '{0} {1}'.format(src, dest)

                instructions.append(' \\\n'.join(c))

            return '\n'.join(instructions)

        elif hpccm.config.g_ctype == container_type.SINGULARITY:
            # If any of the files are being staged in /tmp or /var/tmp,
            # issue a warning or error depending on the Singularity
            # version.
            # https://github.com/NVIDIA/hpc-container-maker/issues/345
            if (not self.__from and
                any(f['dest'].startswith(('/var/tmp', '/tmp')) for f in files)):
                msg = 'Singularity 3.6 and later no longer allow a temporary directory to be used to stage files into the container image.  Modify the recipe or, in many cases, use --working-directory or hpccm.config.set_working_directory() to specify another location.'
                if hpccm.config.g_singularity_version >= StrictVersion('3.6'):
                    raise RuntimeError(msg)
                else:
                    logging.warning(msg)

            # Format:
            # %files
            #     src1 dest
            #     src2 dest
            #     src3 dest
            # %files
            #     src1 dest1
            #     src2 dest2
            #     src3 dest3
            section = '%files'
            if (self.__from and
                hpccm.config.g_singularity_version >= StrictVersion('3.2')):
                section = section + ' from {}'.format(self.__from)

            if self._app:
                # SCIF appfiles does not support "from"
                section = '%appfiles {0}'.format(self._app)

            # Singularity will error if the destination does not
            # already exist in the container.  The workarounds are to
            # either 1) prior to copying the files, create the
            # destination directories with %setup or 2) copy the files
            # to a path guaranteed to exist, "/", and then move them
            # later with %post.  Option 1 is the "pre" approach,
            # option 2 is the "post" approach.
            flat_files = []
            post = [] # post actions if _post is enabled
            pre = [] # pre actions if _mkdir is enabled
            for pair in files:
                dest = pair['dest']
                src = pair['src']

                if self._post:
                    dest = '/'

                if isinstance(src, list):
                    for s in src:
                        flat_files.append('    {0} {1}'.format(s, dest))

                        if self._post:
                            post.append('    mv /{0} {1}'.format(posixpath.basename(s), posixpath.join(pair['dest'], s)))
                    if (self._mkdir and
                        posixpath.dirname(dest) != '/' and
                        posixpath.basename(dest) != dest):
                        # When multiple files are to be copied to the
                        # same destination, assume the destination is
                        # a directory
                        pre.append('    mkdir -p ${{SINGULARITY_ROOTFS}}{0}'.format(dest))
                else:
                    flat_files.append('    {0} {1}'.format(src, dest))
                    if (self._mkdir and
                        posixpath.dirname(dest) != '/' and
                        posixpath.basename(dest) != dest):
                        # When a single file is to be copied to a
                        # destination, assume the destination is a
                        # file.
                        pre.append('    mkdir -p ${{SINGULARITY_ROOTFS}}{0}'.format(posixpath.dirname(dest)))
                    elif self._post:
                        post.append('    mv /{0} {1}'.format(posixpath.basename(src), pair['dest']))

            s = ''
            if pre:
                s += '%setup\n' + '\n'.join(pre) + '\n'
            s += section + '\n' + '\n'.join(flat_files)
            if post:
                s += '\n%post\n' + '\n'.join(post)

            return s

        elif hpccm.config.g_ctype == container_type.BASH:
            logging.warning('copy primitive does not map into bash')
            return ''
        else:
            raise RuntimeError('Unknown container type')

    def merge(self, lst, _app=None):
        """Merge one or more instances of the primitive into a single
        instance.  Due to conflicts or option differences the merged
        primitive may not be exact merger.

        """

        if not lst: # pragma: nocover
            raise RuntimeError('no items provided to merge')

        files = {}
        for item in lst:
            if not item.__class__.__name__ == 'copy': # pragma: nocover
                logging.warning('item is not the correct type, skipping...')
                continue

            if item._copy__files:
                files.update(item._copy__files)
            elif isinstance(item._copy__src, list):
                # Build a files dictionary from src / dest options.
                # src is a list.
                for s in item._copy__src:
                    files.update({s: item._copy__dest})
            else:
                # Build a files dictionary from src / dest options.
                files.update({item._copy__src: item._copy__dest})

        return copy(files=files, _app=_app)



# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods
# pylint: disable=too-many-instance-attributes

"""XPMEM building block"""

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

import posixpath

import hpccm.templates.envvars
import hpccm.templates.ldconfig

from hpccm.building_blocks.base import bb_base
from hpccm.building_blocks.generic_autotools import generic_autotools
from hpccm.building_blocks.packages import packages
from hpccm.primitives.comment import comment

class xpmem(bb_base, hpccm.templates.envvars, hpccm.templates.ldconfig):
    """The `xpmem` building block builds and installs the user space
    library from the [XPMEM](https://gitlab.com/hjelmn/xpmem)
    component.

    # Parameters

    annotate: Boolean flag to specify whether to include annotations
    (labels).  The default is False.

    branch: The branch of XPMEM to use.  The default value is
    `master`.

    configure_opts: List of options to pass to `configure`.  The
    default values are `--disable-kernel-module`.

    disable_FEATURE: Flags to control disabling features when
    configuring.  For instance, `disable_foo=True` maps to
    `--disable-foo`.  Underscores in the parameter name are converted
    to dashes.

    enable_FEATURE[=ARG]: Flags to control enabling features when
    configuring.  For instance, `enable_foo=True` maps to
    `--enable-foo` and `enable_foo='yes'` maps to `--enable-foo=yes`.
    Underscores in the parameter name are converted to dashes.

    environment: Boolean flag to specify whether the environment
    (`CPATH`, `LD_LIBRARY_PATH` and `LIBRARY_PATH`) should be modified
    to include XPMEM. The default is True.

    ldconfig: Boolean flag to specify whether the XPMEM library
    directory should be added dynamic linker cache.  If False, then
    `LD_LIBRARY_PATH` is modified to include the XPMEM library
    directory. The default value is False.

    ospackages: List of OS packages to install prior to configuring
    and building.  The default value are `autoconf`, `automake`,
    `ca-certificates`, `file, `git`, `libtool`, and `make`.

    prefix: The top level install location.  The default value is
    `/usr/local/xpmem`.

    toolchain: The toolchain object.  This should be used if
    non-default compilers or other toolchain options are needed.  The
    default is empty.

    with_PACKAGE[=ARG]: Flags to control optional packages when
    configuring.  For instance, `with_foo=True` maps to `--with-foo`
    and `with_foo='/usr/local/foo'` maps to
    `--with-foo=/usr/local/foo`.  Underscores in the parameter name
    are converted to dashes.

    without_PACKAGE: Flags to control optional packages when
    configuring.  For instance `without_foo=True` maps to
    `--without-foo`.  Underscores in the parameter name are converted
    to dashes.

    # Examples

    ```python
    xpmem(prefix='/opt/xpmem', branch='master')
    ```

    """

    def __init__(self, **kwargs):
        """Initialize building block"""

        super(xpmem, self).__init__(**kwargs)

        # Parameters
        self.__branch = kwargs.pop('branch', 'master')
        self.__configure_opts = kwargs.pop('configure_opts',
                                           ['--disable-kernel-module'])
        self.__ospackages = kwargs.pop('ospackages', ['autoconf', 'automake',
                                                      'ca-certificates',
                                                      'file', 'git',
                                                      'libtool', 'make'])
        self.__prefix = kwargs.pop('prefix', '/usr/local/xpmem')
        self.__repository = kwargs.pop('repository',
                                       'https://gitlab.com/hjelmn/xpmem.git')

        # Setup the environment variables
        self.environment_variables['CPATH'] = '{}:$CPATH'.format(
            posixpath.join(self.__prefix, 'include'))
        self.environment_variables['LIBRARY_PATH'] = '{}:$LIBRARY_PATH'.format(
            posixpath.join(self.__prefix, 'lib'))
        if not self.ldconfig:
            self.environment_variables['LD_LIBRARY_PATH'] = '{}:$LD_LIBRARY_PATH'.format(posixpath.join(self.__prefix, 'lib'))

        # Setup build configuration
        self.__bb = generic_autotools(
            base_annotation=self.__class__.__name__,
            branch=self.__branch,
            comment=False,
            configure_opts=self.__configure_opts,
            devel_environment=self.environment_variables,
            preconfigure=['autoreconf --install'],
            prefix=self.__prefix,
            repository=self.__repository,
            runtime_environment=self.environment_variables,
            **kwargs)

        # Container instructions
        self += comment('XPMEM branch {}'.format(self.__branch))
        self += packages(ospackages=self.__ospackages)
        self += self.__bb

    def runtime(self, _from='0'):
        """Generate the set of instructions to install the runtime specific
        components from a build in a previous stage.

        # Examples

        ```python
        x = xpmem(...)
        Stage0 += x
        Stage1 += x.runtime()
        ```
        """
        self.rt += comment('XPMEM')
        self.rt += self.__bb.runtime(_from=_from)
        return str(self.rt)



# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods, bad-continuation

"""Test cases for the building block base class"""

from __future__ import unicode_literals
from __future__ import print_function

import logging # pylint: disable=unused-import
import unittest

from helpers import centos, docker, ubuntu

from hpccm.building_blocks.base import bb_base
from hpccm.primitives import shell

class Test_bb_base(unittest.TestCase):
    def setUp(self):
        """Disable logging output messages"""
        logging.disable(logging.ERROR)

    @ubuntu
    @docker
    def test_defaults(self):
        """Default building block base class"""
        b = bb_base()
        self.assertEqual(str(b), '')

    @ubuntu
    @docker
    def test_instruction_manipulations(self):
        """Instruction manipulations"""
        b = bb_base()

        # Append instructions
        b += shell(commands=['echo a'])
        # Append directly to "private" class variable (not recommended)
        b._bb_instructions__instructions_bb.append(shell(commands=['echo b']))
        self.assertEqual(len(b), 2)
        self.assertEqual(str(b), 'RUN echo a\nRUN echo b')

        # Direct element access
        self.assertEqual(str(b[0]), 'RUN echo a')
        self.assertEqual(str(b[1]), 'RUN echo b')

        # Iterators
        i = iter(b)
        self.assertEqual(str(next(i)), 'RUN echo a')
        self.assertEqual(str(next(i)), 'RUN echo b')

        # Insertion, using "private" class variable (not recommended)
        b._bb_instructions__instructions_bb.insert(0, shell(commands=['echo c']))
        self.assertEqual(len(b), 3)
        self.assertEqual(str(b), 'RUN echo c\nRUN echo a\nRUN echo b')

        # Deletion (not allowed)
        with self.assertRaises(TypeError):
            del(b[1])

        # Deletion via "private" class variable (not recommended)
        del(b._bb_instructions__instructions_bb[1])
        self.assertEqual(len(b), 2)
        self.assertEqual(str(b), 'RUN echo c\nRUN echo b')

    @ubuntu
    @docker
    def test_runtime_instructions(self):
        """Instruction manipulations"""
        b = bb_base()

        # A non-runtime instruction that should not appear
        b += shell(commands=['echo a'])

        b.rt += shell(commands=['echo r1'])
        b.rt += shell(commands=['echo r2'])
        self.assertEqual(len(b.rt), 2)
        self.assertEqual(str(b.rt), 'RUN echo r1\nRUN echo r2')



# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods, bad-continuation

"""Test cases for the nccl module"""

from __future__ import unicode_literals
from __future__ import print_function

import logging # pylint: disable=unused-import
import unittest

from helpers import centos, centos8, docker, ppc64le, ubuntu, ubuntu18, x86_64

from hpccm.building_blocks.nccl import nccl

class Test_nccl(unittest.TestCase):
    def setUp(self):
        """Disable logging output messages"""
        logging.disable(logging.ERROR)

    @x86_64
    @ubuntu
    @docker
    def test_defaults_ubuntu(self):
        """nccl defaults"""
        n = nccl()
        self.assertEqual(str(n),
r'''# NCCL 2.7.8-1
RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        apt-transport-https \
        ca-certificates \
        gnupg \
        wget && \
    rm -rf /var/lib/apt/lists/*
RUN wget -qO - https://developer.download.nvidia.com/compute/machine-learning/repos/ubuntu1604/x86_64/7fa2af80.pub | apt-key add - && \
    echo "deb https://developer.download.nvidia.com/compute/machine-learning/repos/ubuntu1604/x86_64 /" >> /etc/apt/sources.list.d/hpccm.list && \
    apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        libnccl-dev=2.7.8-1+cuda11.0 \
        libnccl2=2.7.8-1+cuda11.0 && \
    rm -rf /var/lib/apt/lists/*''')

    @x86_64
    @ubuntu18
    @docker
    def test_defaults_ubuntu18(self):
        """nccl defaults"""
        n = nccl()
        self.assertEqual(str(n),
r'''# NCCL 2.7.8-1
RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        apt-transport-https \
        ca-certificates \
        gnupg \
        wget && \
    rm -rf /var/lib/apt/lists/*
RUN wget -qO - https://developer.download.nvidia.com/compute/machine-learning/repos/ubuntu1804/x86_64/7fa2af80.pub | apt-key add - && \
    echo "deb https://developer.download.nvidia.com/compute/machine-learning/repos/ubuntu1804/x86_64 /" >> /etc/apt/sources.list.d/hpccm.list && \
    apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        libnccl-dev=2.7.8-1+cuda11.0 \
        libnccl2=2.7.8-1+cuda11.0 && \
    rm -rf /var/lib/apt/lists/*''')

    @ppc64le
    @ubuntu
    @docker
    def test_ubuntu_ppc64le(self):
        """nccl ppc64le"""
        n = nccl(cuda=9.2, version='2.4.8-1')
        self.assertEqual(str(n),
r'''# NCCL 2.4.8-1
RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        apt-transport-https \
        ca-certificates \
        gnupg \
        wget && \
    rm -rf /var/lib/apt/lists/*
RUN wget -qO - https://developer.download.nvidia.com/compute/machine-learning/repos/ubuntu1604/ppc64el/7fa2af80.pub | apt-key add - && \
    echo "deb https://developer.download.nvidia.com/compute/machine-learning/repos/ubuntu1604/ppc64el /" >> /etc/apt/sources.list.d/hpccm.list && \
    apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        libnccl-dev=2.4.8-1+cuda9.2 \
        libnccl2=2.4.8-1+cuda9.2 && \
    rm -rf /var/lib/apt/lists/*''')

    @x86_64
    @ubuntu
    @docker
    def test_build_ubuntu(self):
        """nccl build"""
        n = nccl(build=True)
        self.assertEqual(str(n),
r'''# NCCL
RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        make \
        wget && \
    rm -rf /var/lib/apt/lists/*
RUN mkdir -p /var/tmp && wget -q -nc --no-check-certificate -P /var/tmp https://github.com/NVIDIA/nccl/archive/v2.7.8-1.tar.gz && \
    mkdir -p /var/tmp && tar -x -f /var/tmp/v2.7.8-1.tar.gz -C /var/tmp -z && \
    cd /var/tmp/nccl-2.7.8-1 && \
    PREFIX=/usr/local/nccl make -j$(nproc) install && \
    rm -rf /var/tmp/nccl-2.7.8-1 /var/tmp/v2.7.8-1.tar.gz
ENV CPATH=/usr/local/nccl/include:$CPATH \
    LD_LIBRARY_PATH=/usr/local/nccl/lib:$LD_LIBRARY_PATH \
    LIBRARY_PATH=/usr/local/nccl/lib:$LIBRARY_PATH \
    PATH=/usr/local/nccl/bin:$PATH''')

    @x86_64
    @centos
    @docker
    def test_defaults_centos(self):
        """nccl defaults"""
        n = nccl()
        self.assertEqual(str(n),
r'''# NCCL 2.7.8-1
RUN rpm --import https://developer.download.nvidia.com/compute/machine-learning/repos/rhel7/x86_64/7fa2af80.pub && \
    yum install -y yum-utils && \
    yum-config-manager --add-repo https://developer.download.nvidia.com/compute/machine-learning/repos/rhel7/x86_64 && \
    yum install -y \
        libnccl-2.7.8-1+cuda11.0 \
        libnccl-devel-2.7.8-1+cuda11.0 && \
    rm -rf /var/cache/yum/*''')

    @x86_64
    @centos8
    @docker
    def test_defaults_centos8(self):
        """nccl defaults"""
        n = nccl()
        self.assertEqual(str(n),
r'''# NCCL 2.7.8-1
RUN rpm --import https://developer.download.nvidia.com/compute/machine-learning/repos/rhel8/x86_64/7fa2af80.pub && \
    yum install -y dnf-utils && \
    yum-config-manager --add-repo https://developer.download.nvidia.com/compute/machine-learning/repos/rhel8/x86_64 && \
    yum install -y \
        libnccl-2.7.8-1+cuda11.0 \
        libnccl-devel-2.7.8-1+cuda11.0 && \
    rm -rf /var/cache/yum/*''')

    @x86_64
    @ubuntu
    @docker
    def test_build_repo_ubuntu(self):
        """nccl build from git"""
        n = nccl(build=True, make_variables={'CUDA_HOME': '/usr/local/cuda'},
                 repository=True)
        self.assertEqual(str(n),
r'''# NCCL
RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        git \
        make \
        wget && \
    rm -rf /var/lib/apt/lists/*
RUN mkdir -p /var/tmp && cd /var/tmp && git clone --depth=1 https://github.com/NVIDIA/nccl.git nccl && cd - && \
    cd /var/tmp/nccl && \
    CUDA_HOME=/usr/local/cuda PREFIX=/usr/local/nccl make -j$(nproc) install && \
    rm -rf /var/tmp/nccl
ENV CPATH=/usr/local/nccl/include:$CPATH \
    LD_LIBRARY_PATH=/usr/local/nccl/lib:$LD_LIBRARY_PATH \
    LIBRARY_PATH=/usr/local/nccl/lib:$LIBRARY_PATH \
    PATH=/usr/local/nccl/bin:$PATH''')

    @centos
    @docker
    def test_build_centos(self):
        """nccl build"""
        n = nccl(build=True, version='2.7.6-1')
        self.assertEqual(str(n),
r'''# NCCL
RUN yum install -y \
        make \
        wget \
        which && \
    rm -rf /var/cache/yum/*
RUN mkdir -p /var/tmp && wget -q -nc --no-check-certificate -P /var/tmp https://github.com/NVIDIA/nccl/archive/v2.7.6-1.tar.gz && \
    mkdir -p /var/tmp && tar -x -f /var/tmp/v2.7.6-1.tar.gz -C /var/tmp -z && \
    cd /var/tmp/nccl-2.7.6-1 && \
    PREFIX=/usr/local/nccl make -j$(nproc) install && \
    rm -rf /var/tmp/nccl-2.7.6-1 /var/tmp/v2.7.6-1.tar.gz
ENV CPATH=/usr/local/nccl/include:$CPATH \
    LD_LIBRARY_PATH=/usr/local/nccl/lib:$LD_LIBRARY_PATH \
    LIBRARY_PATH=/usr/local/nccl/lib:$LIBRARY_PATH \
    PATH=/usr/local/nccl/bin:$PATH''')

    @x86_64
    @ubuntu
    @docker
    def test_runtime(self):
        """Runtime"""
        n = nccl()
        r = n.runtime()
        self.assertEqual(r,
r'''# NCCL
RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        apt-transport-https \
        ca-certificates \
        gnupg \
        wget && \
    rm -rf /var/lib/apt/lists/*
RUN wget -qO - https://developer.download.nvidia.com/compute/machine-learning/repos/ubuntu1604/x86_64/7fa2af80.pub | apt-key add - && \
    echo "deb https://developer.download.nvidia.com/compute/machine-learning/repos/ubuntu1604/x86_64 /" >> /etc/apt/sources.list.d/hpccm.list && \
    apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        libnccl2=2.7.8-1+cuda11.0 && \
    rm -rf /var/lib/apt/lists/*''')

    @x86_64
    @ubuntu
    @docker
    def test_build_runtime(self):
        """Runtime"""
        n = nccl(build=True)
        r = n.runtime()
        self.assertEqual(r,
r'''# NCCL
COPY --from=0 /usr/local/nccl /usr/local/nccl
ENV CPATH=/usr/local/nccl/include:$CPATH \
    LD_LIBRARY_PATH=/usr/local/nccl/lib:$LD_LIBRARY_PATH \
    LIBRARY_PATH=/usr/local/nccl/lib:$LIBRARY_PATH \
    PATH=/usr/local/nccl/bin:$PATH''')



# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods

"""Unit testing helpers"""

from __future__ import unicode_literals
from __future__ import print_function

from distutils.version import StrictVersion
import logging # pylint: disable=unused-import

import hpccm.config

from hpccm.common import container_type, cpu_arch, linux_distro

def aarch64(function):
    """Decorator to set the CPU architecture to aarch64"""
    def wrapper(*args, **kwargs):
        hpccm.config.g_cpu_arch = cpu_arch.AARCH64
        return function(*args, **kwargs)

    return wrapper

def bash(function):
    """Decorator to set the global container type to bash"""
    def wrapper(*args, **kwargs):
        hpccm.config.g_ctype = container_type.BASH
        return function(*args, **kwargs)

    return wrapper

def broadwell(function):
    """Decorator to set the CPU type to Broadwell"""
    def wrapper(*args, **kwargs):
        hpccm.config.g_cpu_target = 'broadwell'
        return function(*args, **kwargs)

    return wrapper

def centos(function):
    """Decorator to set the Linux distribution to CentOS 7"""
    def wrapper(*args, **kwargs):
        hpccm.config.g_linux_distro = linux_distro.CENTOS
        hpccm.config.g_linux_version = StrictVersion('7.0')
        return function(*args, **kwargs)

    return wrapper

def centos8(function):
    """Decorator to set the Linux distribution to CentOS 8"""
    def wrapper(*args, **kwargs):
        hpccm.config.g_linux_distro = linux_distro.CENTOS
        hpccm.config.g_linux_version = StrictVersion('8.0')
        return function(*args, **kwargs)

    return wrapper

def cpu_target_none(function):
    """Decorator to set the CPU type to None"""
    def wrapper(*args, **kwargs):
        hpccm.config.g_cpu_target = None
        return function(*args, **kwargs)

    return wrapper

def docker(function):
    """Decorator to set the global container type to docker"""
    def wrapper(*args, **kwargs):
        hpccm.config.g_ctype = container_type.DOCKER
        return function(*args, **kwargs)

    return wrapper

def icelake(function):
    """Decorator to set the CPU type to Ice Lake"""
    def wrapper(*args, **kwargs):
        hpccm.config.g_cpu_target = 'icelake'
        return function(*args, **kwargs)

    return wrapper

def invalid_cpu_arch(function):
    """Decorator to set the global CPU architecture to an invalid value"""
    def wrapper(*args, **kwargs):
        hpccm.config.g_cpu_arch = None
        return function(*args, **kwargs)

    return wrapper

def invalid_ctype(function):
    """Decorator to set the global container type to an invalid value"""
    def wrapper(*args, **kwargs):
        hpccm.config.g_ctype = None
        return function(*args, **kwargs)

    return wrapper

def invalid_distro(function):
    """Decorator to set the global Linux distribution to an invalid value"""
    def wrapper(*args, **kwargs):
        hpccm.config.g_linux_distro = None
        return function(*args, **kwargs)

    return wrapper

def ppc64le(function):
    """Decorator to set the CPU architecture to ppc64le"""
    def wrapper(*args, **kwargs):
        hpccm.config.g_cpu_arch = cpu_arch.PPC64LE
        return function(*args, **kwargs)

    return wrapper

def singularity(function):
    """Decorator to set the global container type to singularity"""
    def wrapper(*args, **kwargs):
        hpccm.config.g_ctype = container_type.SINGULARITY
        return function(*args, **kwargs)

    return wrapper

def singularity26(function):
    """Decorator to set the global singularity version"""
    def wrapper(*args, **kwargs):
        hpccm.config.g_ctype = container_type.SINGULARITY
        hpccm.config.g_singularity_version = StrictVersion('2.6')
        return function(*args, **kwargs)

    return wrapper

def singularity32(function):
    """Decorator to set the global singularity version"""
    def wrapper(*args, **kwargs):
        hpccm.config.g_ctype = container_type.SINGULARITY
        hpccm.config.g_singularity_version = StrictVersion('3.2')
        return function(*args, **kwargs)

    return wrapper

def singularity37(function):
    """Decorator to set the global singularity version"""
    def wrapper(*args, **kwargs):
        hpccm.config.g_ctype = container_type.SINGULARITY
        hpccm.config.g_singularity_version = StrictVersion('3.7')
        return function(*args, **kwargs)

    return wrapper

def thunderx2(function):
    """Decorator to set the CPU type to ThunderX2"""
    def wrapper(*args, **kwargs):
        hpccm.config.g_cpu_target = 'thunderx2'
        return function(*args, **kwargs)

    return wrapper

def ubuntu(function):
    """Decorator to set the Linux distribution to Ubuntu 16.04"""
    def wrapper(*args, **kwargs):
        hpccm.config.g_linux_distro = linux_distro.UBUNTU
        hpccm.config.g_linux_version = StrictVersion('16.04')
        return function(*args, **kwargs)

    return wrapper

def ubuntu18(function):
    """Decorator to set the Linux distribution to Ubuntu 18.04"""
    def wrapper(*args, **kwargs):
        hpccm.config.g_linux_distro = linux_distro.UBUNTU
        hpccm.config.g_linux_version = StrictVersion('18.04')
        return function(*args, **kwargs)

    return wrapper

def ubuntu20(function):
    """Decorator to set the Linux distribution to Ubuntu 20.04"""
    def wrapper(*args, **kwargs):
        hpccm.config.g_linux_distro = linux_distro.UBUNTU
        hpccm.config.g_linux_version = StrictVersion('20.04')
        return function(*args, **kwargs)

    return wrapper

def x86_64(function):
    """Decorator to set the CPU architecture to x86_64"""
    def wrapper(*args, **kwargs):
        hpccm.config.g_cpu_arch = cpu_arch.X86_64
        return function(*args, **kwargs)

    return wrapper

def zen2(function):
    """Decorator to set the CPU type to Zen 2"""
    def wrapper(*args, **kwargs):
        hpccm.config.g_cpu_target = 'zen2'
        return function(*args, **kwargs)

    return wrapper



# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from hpccm.version import __version__

from hpccm.base_object import base_object

from hpccm.common import cpu_arch
from hpccm.common import container_type
from hpccm.common import linux_distro

from hpccm.Stage import Stage
from hpccm.recipe import include
from hpccm.recipe import recipe
from hpccm.toolchain import toolchain

import hpccm.building_blocks
import hpccm.templates
import hpccm.primitives

# Templates
# For backwards compatibility with recipes that use "hpccm.git()", etc.
from hpccm.templates.ConfigureMake import ConfigureMake
from hpccm.templates.git import git
from hpccm.templates.rm import rm
from hpccm.templates.sed import sed
from hpccm.templates.tar import tar
from hpccm.templates.wget import wget



# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods

"""Test cases for the rm module"""

from __future__ import unicode_literals
from __future__ import print_function

import logging # pylint: disable=unused-import
import unittest

from hpccm.templates.rm import rm

class Test_rm(unittest.TestCase):
    def setUp(self):
        """Disable logging output messages"""
        logging.disable(logging.ERROR)

    def test_missing_files(self):
        """Missing files"""
        r = rm()
        self.assertEqual(r.cleanup_step(), '')

    def test_basic(self):
        """Basic file removal"""
        r = rm()

        self.assertEqual(r.cleanup_step(items=['a', 'b', 'c']),
                         'rm -rf a b c')



# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods
# pylint: disable=too-many-instance-attributes

"""pip building block"""

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

from distutils.version import StrictVersion
import logging
import posixpath

import hpccm.config
import hpccm.templates.rm

from hpccm.building_blocks.base import bb_base
from hpccm.building_blocks.packages import packages
from hpccm.common import linux_distro
from hpccm.primitives.comment import comment
from hpccm.primitives.copy import copy
from hpccm.primitives.shell import shell

class pip(bb_base, hpccm.templates.rm):
    """The `pip` building block installs Python packages from PyPi.

    # Parameters

    alternatives: Boolean flag to specify whether to configure alternatives for `python` and `pip`.  RHEL-based 8.x distributions do not setup `python` by [default](https://developers.redhat.com/blog/2019/05/07/what-no-python-in-red-hat-enterprise-linux-8/).  The default is False.

    args: List of arguments to pass to pip.  The default is
    `--no-cache-dir`.

    ospackages: List of OS packages to install prior to installing
    PyPi packages.  For Ubuntu, the default values are `python-pip`,
    `python-setuptools`, and `python-wheel` for Python 2.x and
    `python3-pip`, `python3-setuptools`, and `python3-wheel` for
    Python 3.x.  For RHEL-based distributions, the default
    values are `python2-pip` for Python 2.x and `python3-pip` for
    Python 3.x.

    packages: List of PyPi packages to install.  The default is
    an empty list.

    pip: The name of the `pip` tool to use. The default is `pip`.

    requirements: Path to pip requirements file.  The default is
    empty.

    upgrade: Boolean flag to control whether pip itself should be
    upgraded prior to installing any PyPi packages.  The default is
    False.

    # Examples

    ```python
    pip(packages=['hpccm'])
    ```

    ```python
    pip(packages=['hpccm'], pip='pip3')
    ```

    ```python
    pip(requirements='requirements.txt')
    ```

    """

    def __init__(self, **kwargs):
        """Initialize building block"""

        super(pip, self).__init__(**kwargs)

        self.__alternatives = kwargs.get('alternatives', False)
        self.__args = kwargs.get('args', ['--no-cache-dir'])
        self.__epel = False
        self.__ospackages = kwargs.get('ospackages', None)
        self.__packages = kwargs.get('packages', [])
        self.__pip = kwargs.get('pip', 'pip')
        self.__requirements = kwargs.get('requirements', None)
        self.__upgrade = kwargs.get('upgrade', False)
        self.__wd = kwargs.get('wd', hpccm.config.g_wd) # working directory

        self.__debs = [] # Filled in below
        self.__rpms = [] # Filled in below

        if self.__ospackages is None:
            if self.__pip.startswith('pip3'):
                self.__debs.extend(['python3-pip', 'python3-setuptools',
                                    'python3-wheel'])
                self.__rpms.append('python3-pip')
            else:
                self.__debs.extend(['python-pip', 'python-setuptools',
                                    'python-wheel'])
                self.__rpms.append('python2-pip')
                if (hpccm.config.g_linux_distro == linux_distro.CENTOS and
                    hpccm.config.g_linux_version < StrictVersion('8.0')):
                    # python2-pip is an EPEL package in CentOS 7.x
                    self.__epel = True
                elif (hpccm.config.g_linux_distro == linux_distro.UBUNTU and
                      hpccm.config.g_linux_version >= StrictVersion('20.0')):
                    # python-pip is not supported in Ubuntu 20.04
                    logging.warning('pip2 is not supported in Ubuntu 20.04.  Use pip3.')
        elif self.__ospackages:
            self.__debs = self.__ospackages
            self.__rpms = self.__ospackages

        # Fill in container instructions
        self.__instructions()

    def __instructions(self):
        """Fill in container instructions"""

        self += comment('pip')
        if self.__debs or self.__rpms:
            self += packages(apt=self.__debs, epel=self.__epel,
                             yum=self.__rpms)

        if self.__alternatives:
            self += shell(commands=[
                'alternatives --set python /usr/bin/python2',
                'alternatives --install /usr/bin/pip pip /usr/bin/pip2 30'])

        if self.__pip:
            if self.__args:
                self.__pip = '{0} {1}'.format(self.__pip,
                                              ' '.join(self.__args))

            cmds = []

            if self.__upgrade:
                # pip version 21 and later no longer support Python 2
                if self.__pip.startswith('pip3'):
                    cmds.append('{0} install --upgrade pip'.format(self.__pip))
                else:
                    cmds.append('{0} install --upgrade "pip < 21.0"'.format(
                        self.__pip))

            if self.__requirements:
                self += copy(src=self.__requirements,
                             dest=posixpath.join(
                                 self.__wd,
                                 posixpath.basename(self.__requirements)))
                cmds.append('{0} install -r {1}'.format(
                    self.__pip,
                    posixpath.join(self.__wd,
                                   posixpath.basename(self.__requirements))))
                cmds.append(self.cleanup_step(items=[
                    posixpath.join(self.__wd,
                                   posixpath.basename(self.__requirements))]))

            if self.__packages:
                cmds.append('{0} install {1}'.format(self.__pip,
                                                     ' '.join(self.__packages)))
            self += shell(commands=cmds)



# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods
# pylint: disable=too-many-instance-attributes

"""HDF5 building block"""

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

import posixpath
import re
from copy import copy as _copy

import hpccm.config
import hpccm.templates.envvars
import hpccm.templates.ldconfig

from hpccm.building_blocks.base import bb_base
from hpccm.building_blocks.generic_autotools import generic_autotools
from hpccm.building_blocks.packages import packages
from hpccm.common import linux_distro
from hpccm.primitives.comment import comment
from hpccm.toolchain import toolchain

class hdf5(bb_base, hpccm.templates.envvars, hpccm.templates.ldconfig):
    """The `hdf5` building block downloads, configures, builds, and
    installs the [HDF5](http://www.hdfgroup.org) component.  Depending
    on the parameters, the source will be downloaded from the web
    (default) or copied from a source directory in the local build
    context.

    # Parameters

    annotate: Boolean flag to specify whether to include annotations
    (labels).  The default is False.

    check: Boolean flag to specify whether the `make check` step
    should be performed.  The default is False.

    configure_opts: List of options to pass to `configure`.  The
    default values are `--enable-cxx` and `--enable-fortran`.

    directory: Path to the unpackaged source directory relative to the
    local build context.  The default value is empty.  If this is
    defined, the source in the local build context will be used rather
    than downloading the source from the web.

    disable_FEATURE: Flags to control disabling features when
    configuring.  For instance, `disable_foo=True` maps to
    `--disable-foo`.  Underscores in the parameter name are converted
    to dashes.

    enable_FEATURE[=ARG]: Flags to control enabling features when
    configuring.  For instance, `enable_foo=True` maps to
    `--enable-foo` and `enable_foo='yes'` maps to `--enable-foo=yes`.
    Underscores in the parameter name are converted to dashes.

    environment: Boolean flag to specify whether the environment
    (`CPATH`, `LD_LIBRARY_PATH`, `LIBRARY_PATH`, `PATH`, and others)
    should be modified to include HDF5. The default is True.

    ldconfig: Boolean flag to specify whether the HDF5 library
    directory should be added dynamic linker cache.  If False, then
    `LD_LIBRARY_PATH` is modified to include the HDF5 library
    directory. The default value is False.

    ospackages: List of OS packages to install prior to configuring
    and building.  For Ubuntu, the default values are `bzip2`, `file`,
    `make`, `wget`, and `zlib1g-dev`.  For RHEL-based Linux
    distributions the default values are `bzip2`, `file`, `make`,
    `wget` and `zlib-devel`.

    prefix: The top level install location.  The default value is
    `/usr/local/hdf5`.

    toolchain: The toolchain object.  This should be used if
    non-default compilers or other toolchain options are needed.  The
    default is empty.

    version: The version of HDF5 source to download.  This value is
    ignored if `directory` is set.  The default value is `1.12.0`.

    with_PACKAGE[=ARG]: Flags to control optional packages when
    configuring.  For instance, `with_foo=True` maps to `--with-foo`
    and `with_foo='/usr/local/foo'` maps to
    `--with-foo=/usr/local/foo`.  Underscores in the parameter name
    are converted to dashes.

    without_PACKAGE: Flags to control optional packages when
    configuring.  For instance `without_foo=True` maps to
    `--without-foo`.  Underscores in the parameter name are converted
    to dashes.

    # Examples

    ```python
    hdf5(prefix='/opt/hdf5/1.10.1', version='1.10.1')
    ```

    ```python
    hdf5(directory='sources/hdf5-1.10.1')
    ```

    ```python
    n = nvhpc(eula=True)
    hdf5(toolchain=n.toolchain)
    ```

    ```python
    hdf5(check=True, configure_opts=['--enable-cxx', '--enable-fortran',
                                     '--enable-profiling=yes'])
    ```

    """

    def __init__(self, **kwargs):
        """Initialize building block"""

        super(hdf5, self).__init__(**kwargs)

        self.__baseurl = kwargs.pop('baseurl', 'http://www.hdfgroup.org/ftp/HDF5/releases')
        self.__check = kwargs.pop('check', False)
        self.__configure_opts = kwargs.pop('configure_opts',
                                           ['--enable-cxx',
                                            '--enable-fortran'])
        self.__ospackages = kwargs.pop('ospackages', [])
        self.__prefix = kwargs.pop('prefix', '/usr/local/hdf5')
        # Create a copy of the toolchain so that it can be modified
        # without impacting the original
        self.__toolchain = _copy(kwargs.pop('toolchain', toolchain()))
        self.__runtime_ospackages = [] # Filled in by __distro()
        self.__version = kwargs.pop('version', '1.12.0')

        # Set the Linux distribution specific parameters
        self.__distro()

        # Set the download specific parameters
        self.__download()

        # Set the environment variables
        self.environment_variables['CPATH'] = '{}:$CPATH'.format(
            posixpath.join(self.__prefix, 'include'))
        self.environment_variables['HDF5_DIR'] = self.__prefix
        self.environment_variables['LIBRARY_PATH'] = '{}:$LIBRARY_PATH'.format(
            posixpath.join(self.__prefix, 'lib'))
        self.environment_variables['PATH'] = '{}:$PATH'.format(
            posixpath.join(self.__prefix, 'bin'))
        if not self.ldconfig:
            self.environment_variables['LD_LIBRARY_PATH'] = '{}:$LD_LIBRARY_PATH'.format(posixpath.join(self.__prefix, 'lib'))

        # PIC workaround when using the NVIDIA compilers
        if self.__toolchain.FC and re.match('.*nvfortran',
                                            self.__toolchain.FC):
            if not self.__toolchain.FCFLAGS:
                self.__toolchain.FCFLAGS = '-fpic -DPIC'

        # Setup build configuration
        self.__bb = generic_autotools(
            annotations={'version': self.__version},
            base_annotation=self.__class__.__name__,
            check=self.__check,
            configure_opts=self.__configure_opts,
            comment=False,
            devel_environment=self.environment_variables,
            prefix=self.__prefix,
            runtime_environment=self.environment_variables,
            toolchain=self.__toolchain,
            url=self.__url,
            **kwargs)

        # Container instructions
        self += comment('HDF5 version {}'.format(self.__version))
        self += packages(ospackages=self.__ospackages)
        self += self.__bb

    def __distro(self):
        """Based on the Linux distribution, set values accordingly.  A user
        specified value overrides any defaults."""

        if hpccm.config.g_linux_distro == linux_distro.UBUNTU:
            if not self.__ospackages:
                self.__ospackages = ['bzip2', 'file', 'make', 'wget',
                                     'zlib1g-dev']
            self.__runtime_ospackages = ['zlib1g']
        elif hpccm.config.g_linux_distro == linux_distro.CENTOS:
            if not self.__ospackages:
                self.__ospackages = ['bzip2', 'file', 'make', 'wget',
                                     'zlib-devel']
                if self.__check:
                    self.__ospackages.append('diffutils')
            self.__runtime_ospackages = ['zlib']
        else: # pragma: no cover
            raise RuntimeError('Unknown Linux distribution')

    def __download(self):
        """Construct the series of shell commands, i.e., fill in
           self.__commands"""

        # The download URL has the format contains vMAJOR.MINOR in the
        # path and the tarball contains MAJOR.MINOR.REVISION, so pull
        # apart the full version to get the MAJOR and MINOR components.
        match = re.match(r'(?P<major>\d+)\.(?P<minor>\d+)', self.__version)
        major_minor = '{0}.{1}'.format(match.groupdict()['major'],
                                       match.groupdict()['minor'])
        tarball = 'hdf5-{}.tar.bz2'.format(self.__version)
        self.__url = '{0}/hdf5-{1}/hdf5-{2}/src/{3}'.format(
            self.__baseurl, major_minor, self.__version, tarball)

    def runtime(self, _from='0'):
        """Generate the set of instructions to install the runtime specific
        components from a build in a previous stage.

        # Examples

        ```python
        h = hdf5(...)
        Stage0 += h
        Stage1 += h.runtime()
        ```
        """
        self.rt += comment('HDF5')
        self.rt += packages(ospackages=self.__runtime_ospackages)
        self.rt += self.__bb.runtime(_from=_from)
        return str(self.rt)



# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods, bad-continuation

"""Test cases for the recipe module"""

from __future__ import unicode_literals
from __future__ import print_function

import logging # pylint: disable=unused-import
import os
import unittest

from hpccm.common import container_type
from hpccm.recipe import recipe

class Test_recipe(unittest.TestCase):
    def setUp(self):
        """Disable logging output messages"""
        logging.disable(logging.ERROR)

    def test_no_arguments(self):
        """No arguments"""

        with self.assertRaises(TypeError):
            recipe()

    def test_bad_recipe(self):
        """Bad (invalid) recipe file"""
        path = os.path.dirname(__file__)
        rf = os.path.join(path, 'bad_recipe.py')
        with self.assertRaises(SystemExit):
            recipe(rf)

    def test_raise_exceptions(self):
        """Bad (invalid) recipe file with raise exceptions enabled"""
        path = os.path.dirname(__file__)
        rf = os.path.join(path, 'bad_recipe.py')
        with self.assertRaises(SyntaxError):
            recipe(rf, raise_exceptions=True)

    def test_basic_example(self):
        """Basic example"""
        path = os.path.dirname(__file__)
        rf = os.path.join(path, '..', 'recipes', 'examples', 'basic.py')
        r = recipe(rf)
        self.assertEqual(r.strip(),
r'''FROM ubuntu:16.04

RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        g++ \
        gcc \
        gfortran && \
    rm -rf /var/lib/apt/lists/*''')

    def test_basic_example_singularity(self):
        """ctype option"""
        path = os.path.dirname(__file__)
        rf = os.path.join(path, '..', 'recipes', 'examples', 'basic.py')
        r = recipe(rf, ctype=container_type.SINGULARITY)
        self.assertEqual(r.strip(),
r'''BootStrap: docker
From: ubuntu:16.04
%post
    . /.singularity.d/env/10-docker*.sh

%post
    apt-get update -y
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        g++ \
        gcc \
        gfortran
    rm -rf /var/lib/apt/lists/*''')

    def test_multistage_example_singlestage(self):
        """Single_stage option"""
        path = os.path.dirname(__file__)
        rf = os.path.join(path, '..', 'recipes', 'examples', 'multistage.py')
        r = recipe(rf, single_stage=True)
        self.assertEqual(r.strip(),
r'''FROM nvcr.io/nvidia/cuda:9.0-devel-ubuntu16.04 AS devel

# GNU compiler
RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        g++ \
        gcc \
        gfortran && \
    rm -rf /var/lib/apt/lists/*

# FFTW version 3.3.8
RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        file \
        make \
        wget && \
    rm -rf /var/lib/apt/lists/*
RUN mkdir -p /var/tmp && wget -q -nc --no-check-certificate -P /var/tmp ftp://ftp.fftw.org/pub/fftw/fftw-3.3.8.tar.gz && \
    mkdir -p /var/tmp && tar -x -f /var/tmp/fftw-3.3.8.tar.gz -C /var/tmp -z && \
    cd /var/tmp/fftw-3.3.8 &&   ./configure --prefix=/usr/local/fftw --enable-openmp --enable-shared --enable-sse2 --enable-threads && \
    make -j$(nproc) && \
    make -j$(nproc) install && \
    rm -rf /var/tmp/fftw-3.3.8 /var/tmp/fftw-3.3.8.tar.gz
ENV LD_LIBRARY_PATH=/usr/local/fftw/lib:$LD_LIBRARY_PATH''')

    def test_multistage_example_singularity26(self):
        """Multi-stage recipe with Singularity container type"""
        path = os.path.dirname(__file__)
        rf = os.path.join(path, '..', 'recipes', 'examples', 'multistage.py')
        r = recipe(rf, ctype=container_type.SINGULARITY,
                   singularity_version='2.6')
        self.assertEqual(r.strip(),
r'''BootStrap: docker
From: nvcr.io/nvidia/cuda:9.0-devel-ubuntu16.04
%post
    . /.singularity.d/env/10-docker*.sh

# GNU compiler
%post
    apt-get update -y
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        g++ \
        gcc \
        gfortran
    rm -rf /var/lib/apt/lists/*

# FFTW version 3.3.8
%post
    apt-get update -y
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        file \
        make \
        wget
    rm -rf /var/lib/apt/lists/*
%post
    cd /
    mkdir -p /var/tmp && wget -q -nc --no-check-certificate -P /var/tmp ftp://ftp.fftw.org/pub/fftw/fftw-3.3.8.tar.gz
    mkdir -p /var/tmp && tar -x -f /var/tmp/fftw-3.3.8.tar.gz -C /var/tmp -z
    cd /var/tmp/fftw-3.3.8 &&   ./configure --prefix=/usr/local/fftw --enable-openmp --enable-shared --enable-sse2 --enable-threads
    make -j$(nproc)
    make -j$(nproc) install
    rm -rf /var/tmp/fftw-3.3.8 /var/tmp/fftw-3.3.8.tar.gz
%environment
    export LD_LIBRARY_PATH=/usr/local/fftw/lib:$LD_LIBRARY_PATH
%post
    export LD_LIBRARY_PATH=/usr/local/fftw/lib:$LD_LIBRARY_PATH''')

    def test_multistage_example_singularity32(self):
        """Multi-stage recipe with Singularity container type"""
        path = os.path.dirname(__file__)
        rf = os.path.join(path, '..', 'recipes', 'examples', 'multistage.py')
        r = recipe(rf, ctype=container_type.SINGULARITY,
                   singularity_version='3.2')
        self.maxDiff = None
        self.assertEqual(r.strip(),
r'''# NOTE: this definition file depends on features only available in
# Singularity 3.2 and later.
BootStrap: docker
From: nvcr.io/nvidia/cuda:9.0-devel-ubuntu16.04
Stage: devel
%post
    . /.singularity.d/env/10-docker*.sh

# GNU compiler
%post
    apt-get update -y
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        g++ \
        gcc \
        gfortran
    rm -rf /var/lib/apt/lists/*

# FFTW version 3.3.8
%post
    apt-get update -y
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        file \
        make \
        wget
    rm -rf /var/lib/apt/lists/*
%post
    cd /
    mkdir -p /var/tmp && wget -q -nc --no-check-certificate -P /var/tmp ftp://ftp.fftw.org/pub/fftw/fftw-3.3.8.tar.gz
    mkdir -p /var/tmp && tar -x -f /var/tmp/fftw-3.3.8.tar.gz -C /var/tmp -z
    cd /var/tmp/fftw-3.3.8 &&   ./configure --prefix=/usr/local/fftw --enable-openmp --enable-shared --enable-sse2 --enable-threads
    make -j$(nproc)
    make -j$(nproc) install
    rm -rf /var/tmp/fftw-3.3.8 /var/tmp/fftw-3.3.8.tar.gz
%environment
    export LD_LIBRARY_PATH=/usr/local/fftw/lib:$LD_LIBRARY_PATH
%post
    export LD_LIBRARY_PATH=/usr/local/fftw/lib:$LD_LIBRARY_PATH

BootStrap: docker
From: nvcr.io/nvidia/cuda:9.0-runtime-ubuntu16.04
%post
    . /.singularity.d/env/10-docker*.sh

# GNU compiler runtime
%post
    apt-get update -y
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        libgfortran3 \
        libgomp1
    rm -rf /var/lib/apt/lists/*

# FFTW
%files from devel
    /usr/local/fftw /usr/local/fftw
%environment
    export LD_LIBRARY_PATH=/usr/local/fftw/lib:$LD_LIBRARY_PATH
%post
    export LD_LIBRARY_PATH=/usr/local/fftw/lib:$LD_LIBRARY_PATH''')

    def test_userarg_example(self):
        """userarg option"""
        path = os.path.dirname(__file__)
        rf = os.path.join(path, '..', 'recipes', 'examples', 'userargs.py')
        r = recipe(rf, userarg={'cuda': '9.0', 'ompi': '2.1.2'})
        self.assertEqual(r.strip(),
r'''FROM nvcr.io/nvidia/cuda:9.0-devel-ubuntu16.04

# OpenMPI version 2.1.2
RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        bzip2 \
        file \
        hwloc \
        libnuma-dev \
        make \
        openssh-client \
        perl \
        tar \
        wget && \
    rm -rf /var/lib/apt/lists/*
RUN mkdir -p /var/tmp && wget -q -nc --no-check-certificate -P /var/tmp https://www.open-mpi.org/software/ompi/v2.1/downloads/openmpi-2.1.2.tar.bz2 && \
    mkdir -p /var/tmp && tar -x -f /var/tmp/openmpi-2.1.2.tar.bz2 -C /var/tmp -j && \
    cd /var/tmp/openmpi-2.1.2 &&   ./configure --prefix=/usr/local/openmpi --disable-getpwuid --enable-orterun-prefix-by-default --with-cuda --without-verbs && \
    make -j$(nproc) && \
    make -j$(nproc) install && \
    rm -rf /var/tmp/openmpi-2.1.2 /var/tmp/openmpi-2.1.2.tar.bz2
ENV LD_LIBRARY_PATH=/usr/local/openmpi/lib:$LD_LIBRARY_PATH \
    PATH=/usr/local/openmpi/bin:$PATH''')

    def test_include(self):
        """recipe include"""
        path = os.path.dirname(__file__)
        rf = os.path.join(path, 'include3.py')
        r = recipe(rf)
        self.assertEqual(r.strip(),
r'''FROM ubuntu:16.04

# GNU compiler
RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        g++ \
        gcc \
        gfortran && \
    rm -rf /var/lib/apt/lists/*''')



# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods

"""Test cases for the wget module"""

from __future__ import unicode_literals
from __future__ import print_function

import logging # pylint: disable=unused-import
import unittest

from hpccm.templates.wget import wget

class Test_wget(unittest.TestCase):
    def setUp(self):
        """Disable logging output messages"""
        logging.disable(logging.ERROR)

    def test_missing_url(self):
        """Missing url option"""
        w = wget()
        self.assertEqual(w.download_step(), '')

    def test_basic(self):
        """Basic wget"""
        w = wget()
        self.assertEqual(w.download_step(url='http://mysite.com/foo.tgz'),
                         'mkdir -p /tmp && wget -q -nc --no-check-certificate -P /tmp http://mysite.com/foo.tgz')

    def test_referer(self):
        """wget with referer"""
        w = wget()
        self.assertEqual(w.download_step(url='http://mysite.com/foo.tgz',
                                         referer='http://mysite.com/foo.html'),
                         'mkdir -p /tmp && wget -q -nc --no-check-certificate --referer http://mysite.com/foo.html -P /tmp http://mysite.com/foo.tgz')

    def test_directory(self):
        """wget with non-default output directory"""
        w = wget()
        self.assertEqual(w.download_step(url='http://mysite.com/foo.tgz',
                                         directory='/scratch'),
                         'mkdir -p /scratch && wget -q -nc --no-check-certificate -P /scratch http://mysite.com/foo.tgz')

    def test_outfile(self):
        """wget with non-default output file"""
        w = wget()
        self.assertEqual(w.download_step(url='http://mysite.com/foo.tgz',
                                         outfile='bar.tgz'),
                         'mkdir -p /tmp && wget -q -nc --no-check-certificate -O bar.tgz -P /tmp http://mysite.com/foo.tgz')

    def test_opts(self):
        """wget with non-default command line options"""
        w = wget(opts=['-fast'])
        self.assertEqual(w.download_step(url='http://mysite.com/foo.tgz'),
                         'mkdir -p /tmp && wget -fast -P /tmp http://mysite.com/foo.tgz')



# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods

"""downloader template"""

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

import logging
import posixpath
import re

import hpccm.base_object
import hpccm.config

from hpccm.common import container_type

class downloader(hpccm.base_object):
    """Template for downloading source code"""

    def __init__(self, **kwargs):
        """Initialize template"""

        self.branch = kwargs.get('branch', None)
        self.commit = kwargs.get('commit', None)
        self.package = kwargs.get('package', None)
        self.repository = kwargs.get('repository', None)
        self.src_directory = None
        self.url = kwargs.get('url', None)

        super(downloader, self).__init__(**kwargs)

    def download_step(self, allow_unknown_filetype=True, recursive=False,
                      unpack=True, wd=hpccm.config.g_wd):
        """Get source code"""

        if not self.repository and not self.package and not self.url:
            raise RuntimeError('must specify a package, repository, or a URL')

        if sum([bool(self.package), bool(self.repository),
                bool(self.url)]) > 1:
            raise RuntimeError('must specify exactly one of a package, repository, or a URL')

        # Check if the caller inherits from the annotate template
        annotate = getattr(self, 'add_annotation', None)

        commands = []

        if self.url:
            # Download package
            commands.append(hpccm.templates.wget().download_step(
                url=self.url, directory=wd))

            if unpack:
                commands.append(self.__unpack(
                    self.url, wd,
                    allow_unknown_filetype=allow_unknown_filetype))

            if callable(annotate):
                self.add_annotation('url', self.url)

        elif self.package:
            # Use an already available package
            if unpack:
                commands.append(self.__unpack(
                    self.package, wd,
                    allow_unknown_filetype=allow_unknown_filetype))

            if callable(annotate):
                self.add_annotation('package', self.package)

        elif self.repository:
            # Clone git repository
            commands.append(hpccm.templates.git().clone_step(
                branch=self.branch, commit=self.commit, path=wd,
                recursive=recursive, repository=self.repository))

            # Set directory where to find source
            self.src_directory = posixpath.join(wd, posixpath.splitext(
                posixpath.basename(self.repository))[0])

            # Add annotations
            if callable(annotate):
                self.add_annotation('repository', self.repository)
                if self.branch:
                    self.add_annotation('branch', self.branch)
                if self.commit:
                    self.add_annotation('commit', self.commit)

        if hpccm.config.g_ctype == container_type.DOCKER:
            return ' && \\\n    '.join(commands)
        elif hpccm.config.g_ctype == container_type.SINGULARITY:
            return '\n    '.join(commands)
        elif hpccm.config.g_ctype == container_type.BASH:
            return '\n'.join(commands)
        else:
            raise RuntimeError('Unknown container type')

    def __unpack(self, package, wd, allow_unknown_filetype=True):
        """Unpack package and set source directory"""

        match_tar = re.search(r'(.*)(?:(?:\.tar)|(?:\.tar\.gz)|(?:\.txz)'
                              r'|(?:\.tgz)|(?:\.tar\.bz2)|(?:\.tar\.xz))$',
                              posixpath.basename(package))

        match_zip = re.search(r'(.*)(?:(?:\.zip))$',
                              posixpath.basename(package))

        if match_tar:
            # Set directory where to find source
            self.src_directory = posixpath.join(wd, match_tar.group(1))
            return hpccm.templates.tar().untar_step(
                posixpath.join(wd, posixpath.basename(package)), directory=wd)
        elif match_zip:
            self.src_directory = posixpath.join(wd, match_zip.group(1))
            return hpccm.templates.zipfile().unzip_step(
                posixpath.join(wd, posixpath.basename(package)), directory=wd)
        elif allow_unknown_filetype:
            # Unclear what the file type is.  For instance, this can
            # happen if a site uses a URL redirector and the shortened
            # URL does not include the file extension.  In most cases,
            # tar can figure it out.  However, note that the src
            # directory is set to None since there is no way to infer
            # what the directory structure might be inside the
            # archive.
            logging.warning('unrecognized package format')
            self.src_directory = None
            return hpccm.templates.tar().untar_step(
                posixpath.join(wd, posixpath.basename(package)), directory=wd)
        else:
            raise RuntimeError('unrecognized package format')



# Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods
# pylint: disable=too-many-instance-attributes

"""Kokkos building block"""

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

from distutils.version import StrictVersion

import hpccm.config
import hpccm.templates.downloader
import hpccm.templates.envvars

from hpccm.building_blocks.base import bb_base
from hpccm.building_blocks.generic_cmake import generic_cmake
from hpccm.building_blocks.packages import packages
from hpccm.common import linux_distro
from hpccm.primitives.comment import comment

class kokkos(bb_base, hpccm.templates.downloader, hpccm.templates.envvars):
    """The `kokkos` building block downloads and installs the
    [Kokkos](https://github.com/kokkos/kokkos) component.

    The [CMake](#cmake) building block should be installed prior to
    this building block.

    # Parameters

    annotate: Boolean flag to specify whether to include annotations
    (labels).  The default is False.

    arch: List of target architectures to build. If set adds
    `-DKokkos_ARCH_<value>=ON` to the list of CMake options. The
    default value is `VOLTA70`, i.e., sm_70.  If a CUDA aware build is
    not selected, then a non-default value should be used.

    branch: The git branch to clone.  Only recognized if the
    `repository` parameter is specified.  The default is empty, i.e.,
    use the default branch for the repository.

    check: Boolean flag to specify whether the build should be
    checked.  If True, adds `-DKokkos_ENABLE_TESTS=ON` to the list of
    CMake options. The default is False.

    cmake_opts: List of options to pass to `cmake`.  The default is
    `-DCMAKE_BUILD_TYPE=RELEASE`.

    commit: The git commit to clone.  Only recognized if the
    `repository` parameter is specified.  The default is empty, i.e.,
    use the latest commit on the default branch for the repository.

    cuda: Flag to control whether a CUDA aware build is performed.  If
    True, adds `-DKokkos_ENABLE_CUDA=ON` and
    `-DCMAKE_CXX_COMPILER=$(pwd)/../bin/nvcc_wrapper` to the list of
    CMake options.  The default value is True.

    environment: Boolean flag to specify whether the environment
    (`LD_LIBRARY_PATH` and `PATH`) should be modified to include
    Kokkos. The default is True.

    hwloc: Flag to control whether a hwloc aware build is performed.
    If True, adds `-DKokkos_ENABLE_HWLOC=ON` to the list of CMake
    options. The default value is True.

    ospackages: List of OS packages to install prior to building.  For
    Ubuntu, the default values are `gzip`, `libhwloc-dev`, `make`,
    `tar`, and `wget`.  For RHEL-based Linux distributions the default
    values are `gzip`, `hwloc-devel`, `make`, `tar`, and `wget`.

    prefix: The top level installation location.  The default value
    is `/usr/local/kokkos`.

    repository: The location of the git repository that should be used to build OpenMPI.  If True, then use the default `https://github.com/kokkos/kokkos.git`
    repository.  The default is empty, i.e., use the release package
    specified by `version`.

    url: The location of the tarball that should be used to build
    Kokkos.  The default is empty, i.e., use the release package
    specified by `version`.

    version: The version of Kokkos source to download.  The default
    value is `3.2.00`.

    # Examples

    ```python
    kokkos(prefix='/opt/kokkos/3.1.01', version='3.1.01')
    ```

    """

    def __init__(self, **kwargs):
        """Initialize building block"""

        super(kokkos, self).__init__(**kwargs)

        self.__arch = kwargs.pop('arch', ['VOLTA70'])
        self.__baseurl = kwargs.pop('baseurl',
                                    'https://github.com/kokkos/kokkos/archive')
        self.__check = kwargs.pop('check', False)
        self.__cmake_opts = kwargs.pop('cmake_opts',
                                       ['-DCMAKE_BUILD_TYPE=RELEASE'])
        self.__cuda = kwargs.pop('cuda', True)
        self.__default_repository = 'https://github.com/kokkos/kokkos.git'
        self.__hwloc = kwargs.pop('hwloc', True)
        self.__ospackages = kwargs.pop('ospackages', [])
        self.__powertools = False # enable the CentOS PowerTools repo
        self.__prefix = kwargs.pop('prefix', '/usr/local/kokkos')
        self.__version = kwargs.pop('version', '3.2.00')

        if self.repository:
            self.__directory = ''
        else:
            self.__directory = kwargs.pop('directory',
                                          'kokkos-{}'.format(self.__version))

        # Set the CMake options
        self.__cmake()

        # Set the Linux distribution specific parameters
        self.__distro()

        # Set the download specific parameters
        self.__download()
        kwargs['repository'] = self.repository
        kwargs['url'] = self.url

        # Setup the environment variables
        self.environment_variables['PATH'] = '{}/bin:$PATH'.format(
            self.__prefix)

        # Setup build configuration
        self.__bb = generic_cmake(
            annotations={'version': self.__version},
            base_annotation=self.__class__.__name__,
            cmake_opts=self.__cmake_opts,
            comment=False,
            devel_environment=self.environment_variables,
            directory=self.__directory,
            prefix=self.__prefix,
            runtime_environment=self.environment_variables,
            **kwargs)

        # Container instructions
        self += comment('Kokkos version {}'.format(self.__version))
        self += packages(ospackages=self.__ospackages,
                         powertools=self.__powertools)
        self += self.__bb

    def __cmake(self):
        """Set CMake options based on user input"""

        # Set options
        if self.__arch:
            for arch in self.__arch:
                self.__cmake_opts.append('-DKokkos_ARCH_{}=ON'.format(
                    arch.upper()))

        if self.__check:
            self.__cmake_opts.append('-DKokkos_ENABLE_TESTS=ON')

        if self.__cuda:
            self.__cmake_opts.append('-DKokkos_ENABLE_CUDA=ON')
            self.__cmake_opts.append(
                '-DCMAKE_CXX_COMPILER=$(pwd)/../bin/nvcc_wrapper')
        if self.__hwloc:
            self.__cmake_opts.append('-DKokkos_ENABLE_HWLOC=ON')

    def __distro(self):
        """Based on the Linux distribution, set values accordingly.  A user
        specified value overrides any defaults."""

        if hpccm.config.g_linux_distro == linux_distro.UBUNTU:
            if not self.__ospackages:
                self.__ospackages = ['libhwloc-dev', 'make']

        elif hpccm.config.g_linux_distro == linux_distro.CENTOS:
            if not self.__ospackages:
                self.__ospackages = ['hwloc-devel', 'make']

            if hpccm.config.g_linux_version >= StrictVersion('8.0'):
                # hwloc-devel is in the CentOS powertools repository
                self.__powertools = True
        else: # pragma: no cover
            raise RuntimeError('Unknown Linux distribution')

        if self.repository:
            self.__ospackages.extend(['ca-certificates', 'git'])
        else:
            self.__ospackages.extend(['gzip', 'tar', 'wget'])

    def __download(self):
        """Set download source based on user parameters"""

        # Use the default repository if set to True
        if self.repository is True:
            self.repository = self.__default_repository

        if not self.repository and not self.url:
            self.url='{0}/{1}.tar.gz'.format(self.__baseurl, self.__version)

    def runtime(self, _from='0'):
        """Generate the set of instructions to install the runtime specific
        components from a build in a previous stage.

        # Examples

        ```python
        k = kokkos(...)
        Stage0 += k
        Stage1 += k.runtime()
        ```
        """
        self.rt += comment('Kokkos')
        self.rt += self.__bb.runtime(_from=_from)
        return str(self.rt)



"""This example demonstrates recipe basics.

Usage:
$ hpccm.py --recipe recipes/examples/basic.py --format docker
# hpccm.py --recipe recipes/examples/basic.py --format singularity
"""

# Choose a base image
Stage0 += baseimage(image='ubuntu:16.04')

# Install GNU compilers (upstream)
Stage0 += apt_get(ospackages=['gcc', 'g++', 'gfortran'])



# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, unused-import

"""Container recipe"""

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function
from six import raise_from

from distutils.version import StrictVersion
import logging
import os
import sys

import hpccm

import hpccm.config

from hpccm.common import container_type

from hpccm.Stage import Stage

from hpccm.building_blocks import *
from hpccm.primitives import *

def include(recipe_file, _globals=None, _locals=None, prepend_path=True,
            raise_exceptions=False):
    """Include a recipe file

    # Arguments

    recipe_file: path to a recipe file (required)

    _globals: a dictionary representing the global symbol table

    _locals: a dictionary representing the local symbol table

    prepend_path: If True, prepend the path of the main recipe to the
    recipe_file. If the recipe_file is an absolute path, then the path
    is not prepended regardless of the value of this parameter.

    raise_exceptions: If False, do not print stack traces when an
    exception is raised.  The default value is False.

    """

    if _locals is None:
        # caller's locals
        _locals = sys._getframe(1).f_locals
    if _globals is None:
        # caller's globals
        _globals = sys._getframe(1).f_globals

    # If a recipe file is included from another recipe file, some way
    # is needed to find the included recipe if it specified using a
    # relative path (relative to the including recipe file). Since
    # recipe files are exec'ed, the value of __file__ is this file,
    # not the recipe file. In order to make including recipes in other
    # recipes using relative paths more intuitive, prepend the path of
    # the base recipe file.
    if (prepend_path and hasattr(include, 'prepend_path')
        and not os.path.isabs(recipe_file)):
        recipe_file = os.path.join(include.prepend_path, recipe_file)

    try:
        with open(recipe_file) as f:
            # pylint: disable=exec-used
            exec(compile(f.read(), recipe_file, 'exec'), _globals, _locals)
    except Exception as e:
        if raise_exceptions:
            raise_from(e, e)
        else:
            logging.error(e)
            exit(1)

def recipe(recipe_file, cpu_target=None, ctype=container_type.DOCKER,
           raise_exceptions=False, single_stage=False,
           singularity_version='2.6', userarg=None,
           working_directory='/var/tmp'):
    """Recipe builder

    # Arguments

    recipe_file: path to a recipe file (required).

    cpu_target: A CPU microarchitecture string recognized by archspec.

    ctype: Enum representing the container specification format.  The
    default is `container_type.DOCKER`.

    raise_exceptions: If False, do not print stack traces when an
    exception is raised.  The default value is False.

    single_stage: If True, only print the first stage of a multi-stage
    recipe.  The default is False.

    singularity_version: Version of the Singularity definition file
    format to use.  Multi-stage support was added in version 3.2, but
    the changes are incompatible with earlier versions of Singularity.
    The default is '2.6'.

    userarg: A dictionary of key / value pairs provided to the recipe
    as the `USERARG` dictionary.

    working_directory: path to use as the working directory in the
    container specification

    """

    # Make user arguments available
    USERARG = {} # pylint: disable=unused-variable
    if userarg:
        USERARG = userarg # alias

    # Consider just 2 stages for the time being
    stages = [Stage(), Stage()]
    Stage0 = stages[0] # alias # pylint: disable=unused-variable
    Stage1 = stages[1] # alias # pylint: disable=unused-variable

    # Set the CPU target
    hpccm.config.set_cpu_target(cpu_target)

    # Set the global container type
    hpccm.config.g_ctype = ctype

    # Set the global Singularity version
    hpccm.config.g_singularity_version = StrictVersion(singularity_version)

    # Set the global working directory
    hpccm.config.g_wd = working_directory

    # Any included recipes that are specified using relative paths will
    # need to prepend the path to the main recipe in order to be found.
    # Save the path to the main recipe.
    include.prepend_path = os.path.dirname(recipe_file)

    # Load in the recipe file
    include(recipe_file, _locals=locals(), _globals=globals(),
            prepend_path=False, raise_exceptions=raise_exceptions)

    # Only process the first stage of a recipe
    if single_stage:
        del stages[1:]
    elif len(Stage1) > 0:
        if (ctype == container_type.SINGULARITY and
            hpccm.config.g_singularity_version < StrictVersion('3.2')):
            # Singularity prior to version 3.2 did not support
            # multi-stage builds.  If the Singularity version is not
            # sufficient to support multi-stage, provide advice to
            # specify a sufficient Singularity version or disable
            # multi-stage.
            logging.warning('This looks like a multi-stage recipe. '
                            'Singularity 3.2 or later is required for '
                            'multi-stage builds.  Use '
                            '--singularity-version=3.2 to enable this '
                            'feature or --single-stage to get rid of this '
                            'warning.  Only processing the first stage...')
            del stages[1:]
        elif ctype == container_type.BASH:
            logging.warning('This looks like a multi-stage recipe, but '
                            'bash does not support multi-stage builds. '
                            'Use --single-stage to get rid of this warning. '
                            'Only processing the first stage...')
            del stages[1:]

    r = []
    for index, stage in enumerate(stages):
        if index >= 1:
            r.append('')
        r.append(str(stage))

    return '\n'.join(r)



# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods, bad-continuation

"""Test cases for the nsight_compute module"""

from __future__ import unicode_literals
from __future__ import print_function

import logging # pylint: disable=unused-import
import unittest

from helpers import aarch64, centos, centos8, docker, ppc64le, ubuntu, ubuntu18, x86_64

from hpccm.building_blocks.nsight_compute import nsight_compute

class Test_nsight_compute(unittest.TestCase):
    def setUp(self):
        """Disable logging output messages"""
        logging.disable(logging.ERROR)

    @x86_64
    @ubuntu
    @docker
    def test_basic_ubuntu(self):
        """Default nsight_compute building block"""
        n = nsight_compute()
        self.assertEqual(str(n),
r'''# NVIDIA Nsight Compute 2020.2.1
RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        apt-transport-https \
        ca-certificates \
        gnupg \
        wget && \
    rm -rf /var/lib/apt/lists/*
RUN wget -qO - https://developer.download.nvidia.com/devtools/repos/ubuntu1604/amd64/nvidia.pub | apt-key add - && \
    echo "deb https://developer.download.nvidia.com/devtools/repos/ubuntu1604/amd64/ /" >> /etc/apt/sources.list.d/hpccm.list && \
    apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        nsight-compute-2020.2.1 && \
    rm -rf /var/lib/apt/lists/*
ENV NV_COMPUTE_PROFILER_DISABLE_STOCK_FILE_DEPLOYMENT=1 \
    PATH=/opt/nvidia/nsight-compute/2020.2.1:$PATH''')

    @x86_64
    @centos8
    @docker
    def test_basic_centos8(self):
        """Default nsight_compute building block"""
        n = nsight_compute()
        self.assertEqual(str(n),
r'''# NVIDIA Nsight Compute 2020.2.1
RUN rpm --import https://developer.download.nvidia.com/devtools/repos/rhel8/x86_64/nvidia.pub && \
    yum install -y dnf-utils && \
    (yum-config-manager --add-repo https://developer.download.nvidia.com/devtools/repos/rhel8/x86_64 || true) && \
    yum install -y \
        nsight-compute-2020.2.1 && \
    rm -rf /var/cache/yum/*
ENV NV_COMPUTE_PROFILER_DISABLE_STOCK_FILE_DEPLOYMENT=1 \
    PATH=/opt/nvidia/nsight-compute/2020.2.1:$PATH''')

    @x86_64
    @ubuntu
    @docker
    def test_version(self):
        """Version option"""
        n = nsight_compute(version='2020.2.1')
        self.assertEqual(str(n),
r'''# NVIDIA Nsight Compute 2020.2.1
RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        apt-transport-https \
        ca-certificates \
        gnupg \
        wget && \
    rm -rf /var/lib/apt/lists/*
RUN wget -qO - https://developer.download.nvidia.com/devtools/repos/ubuntu1604/amd64/nvidia.pub | apt-key add - && \
    echo "deb https://developer.download.nvidia.com/devtools/repos/ubuntu1604/amd64/ /" >> /etc/apt/sources.list.d/hpccm.list && \
    apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        nsight-compute-2020.2.1 && \
    rm -rf /var/lib/apt/lists/*
ENV NV_COMPUTE_PROFILER_DISABLE_STOCK_FILE_DEPLOYMENT=1 \
    PATH=/opt/nvidia/nsight-compute/2020.2.1:$PATH''')

    @ppc64le
    @ubuntu18
    @docker
    def test_ppc64le_ubuntu18(self):
        """Power"""
        n = nsight_compute(version='2020.2.1')
        self.assertEqual(str(n),
r'''# NVIDIA Nsight Compute 2020.2.1
RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        apt-transport-https \
        ca-certificates \
        gnupg \
        wget && \
    rm -rf /var/lib/apt/lists/*
RUN wget -qO - https://developer.download.nvidia.com/devtools/repos/ubuntu1804/ppc64el/nvidia.pub | apt-key add - && \
    echo "deb https://developer.download.nvidia.com/devtools/repos/ubuntu1804/ppc64el/ /" >> /etc/apt/sources.list.d/hpccm.list && \
    apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        nsight-compute-2020.2.1 && \
    rm -rf /var/lib/apt/lists/*
ENV NV_COMPUTE_PROFILER_DISABLE_STOCK_FILE_DEPLOYMENT=1 \
    PATH=/opt/nvidia/nsight-compute/2020.2.1:$PATH''')

    @ppc64le
    @centos
    @docker
    def test_ppc64le_centos(self):
        """Power"""
        n = nsight_compute(version='2020.2.1')
        self.assertEqual(str(n),
r'''# NVIDIA Nsight Compute 2020.2.1
RUN rpm --import https://developer.download.nvidia.com/devtools/repos/rhel7/ppc64le/nvidia.pub && \
    yum install -y yum-utils && \
    (yum-config-manager --add-repo https://developer.download.nvidia.com/devtools/repos/rhel7/ppc64le || true) && \
    yum install -y \
        nsight-compute-2020.2.1 && \
    rm -rf /var/cache/yum/*
ENV NV_COMPUTE_PROFILER_DISABLE_STOCK_FILE_DEPLOYMENT=1 \
    PATH=/opt/nvidia/nsight-compute/2020.2.1:$PATH''')

    @aarch64
    @centos
    @docker
    def test_aarch64_centos(self):
        """Arm"""
        n = nsight_compute(version='2020.2.1')
        self.assertEqual(str(n),
r'''# NVIDIA Nsight Compute 2020.2.1
RUN rpm --import https://developer.download.nvidia.com/devtools/repos/rhel7/arm64/nvidia.pub && \
    yum install -y yum-utils && \
    (yum-config-manager --add-repo https://developer.download.nvidia.com/devtools/repos/rhel7/arm64 || true) && \
    yum install -y \
        nsight-compute-2020.2.1 && \
    rm -rf /var/cache/yum/*
ENV NV_COMPUTE_PROFILER_DISABLE_STOCK_FILE_DEPLOYMENT=1 \
    PATH=/opt/nvidia/nsight-compute/2020.2.1:$PATH''')

    @x86_64
    @ubuntu
    @docker
    def test_runfile(self):
        """Runfile"""
        n = nsight_compute(eula=True,
                           runfile='nsight_compute-linux-x86_64-2020.2.0.18_28964561.run')
        self.assertEqual(str(n),
r'''# NVIDIA Nsight Compute nsight_compute-linux-x86_64-2020.2.0.18_28964561.run
RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        perl \
        wget && \
    rm -rf /var/lib/apt/lists/*
COPY nsight_compute-linux-x86_64-2020.2.0.18_28964561.run /var/tmp/nsight_compute/nsight_compute-linux-x86_64-2020.2.0.18_28964561.run
RUN cd /var/tmp/nsight_compute && \
    sh ./nsight_compute-linux-x86_64-2020.2.0.18_28964561.run --nox11 -- -noprompt -targetpath=/usr/local/NVIDIA-Nsight-Compute && \
    mkdir -p /tmp/var/target && \
    ln -sf /usr/local/NVIDIA-Nsight-Compute/target/* /tmp/var/target && \
    ln -sf /usr/local/NVIDIA-Nsight-Compute/sections /tmp/var/ && \
    chmod -R a+w /tmp/var && \
    rm -rf /var/tmp/nsight_compute /var/tmp/nsight_compute/nsight_compute-linux-x86_64-2020.2.0.18_28964561.run
ENV PATH=/usr/local/NVIDIA-Nsight-Compute:$PATH
ENV NV_COMPUTE_PROFILER_DISABLE_STOCK_FILE_DEPLOYMENT=1''')

    @x86_64
    @ubuntu
    @docker
    def test_basic_ubuntu_url(self):
        """Default nsight_compute building block"""
        n = nsight_compute(eula=True, runfile='https://foo/bar/nsight_compute-linux-x86_64-2020.2.0.18_28964561.run')
        self.assertEqual(str(n),
r'''# NVIDIA Nsight Compute nsight_compute-linux-x86_64-2020.2.0.18_28964561.run
RUN apt-get update -y && \
    DEBIAN_FRONTEND=noninteractive apt-get install -y --no-install-recommends \
        perl \
        wget && \
    rm -rf /var/lib/apt/lists/*
RUN mkdir -p /var/tmp/nsight_compute && wget -q -nc --no-check-certificate -P /var/tmp/nsight_compute https://foo/bar/nsight_compute-linux-x86_64-2020.2.0.18_28964561.run && \
    cd /var/tmp/nsight_compute && \
    sh ./nsight_compute-linux-x86_64-2020.2.0.18_28964561.run --nox11 -- -noprompt -targetpath=/usr/local/NVIDIA-Nsight-Compute && \
    mkdir -p /tmp/var/target && \
    ln -sf /usr/local/NVIDIA-Nsight-Compute/target/* /tmp/var/target && \
    ln -sf /usr/local/NVIDIA-Nsight-Compute/sections /tmp/var/ && \
    chmod -R a+w /tmp/var && \
    rm -rf /var/tmp/nsight_compute /var/tmp/nsight_compute/nsight_compute-linux-x86_64-2020.2.0.18_28964561.run
ENV PATH=/usr/local/NVIDIA-Nsight-Compute:$PATH
ENV NV_COMPUTE_PROFILER_DISABLE_STOCK_FILE_DEPLOYMENT=1''')



# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods
# pylint: disable=too-many-instance-attributes

"""Intel Parallel Studio XE building block"""

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

import logging
import posixpath

import hpccm.config
import hpccm.templates.envvars
import hpccm.templates.rm
import hpccm.templates.sed
import hpccm.templates.tar

from hpccm.building_blocks.base import bb_base
from hpccm.building_blocks.intel_psxe_runtime import intel_psxe_runtime
from hpccm.building_blocks.packages import packages
from hpccm.common import cpu_arch, linux_distro
from hpccm.primitives.comment import comment
from hpccm.primitives.copy import copy
from hpccm.primitives.environment import environment
from hpccm.primitives.shell import shell
from hpccm.toolchain import toolchain

class intel_psxe(bb_base, hpccm.templates.envvars, hpccm.templates.rm,
                 hpccm.templates.sed, hpccm.templates.tar):
    """The `intel_psxe` building block installs [Intel Parallel Studio
    XE](https://software.intel.com/en-us/parallel-studio-xe).

    You must agree to the [Intel End User License Agreement](https://software.intel.com/en-us/articles/end-user-license-agreement)
    to use this building block.

    # Parameters

    components: List of Intel Parallel Studio XE components to
    install.  The default values is `DEFAULTS`.  If only the Intel C++
    and Fortran compilers are desired, then use `intel-icc__x86_64`
    and `intel-ifort__x86_64`.  Please note that the values are not
    consistent between versions; for a list of components, extract
    `pset/mediaconfig.xml` from the tarball and grep for `Abbr`.

    daal: Boolean flag to specify whether the Intel Data Analytics
    Acceleration Library environment should be configured when
    `psxevars` is False.  This flag also controls whether to install
    the corresponding runtime in the `runtime` method.  Note: this
    flag does not control whether the developer environment is
    installed; see `components`.  The default is True.

    environment: Boolean flag to specify whether the environment
    (`LD_LIBRARY_PATH`, `PATH`, and others) should be modified to
    include Intel Parallel Studio XE. `psxevars` has precedence. The
    default is True.

    eula: By setting this value to `True`, you agree to the [Intel End User License Agreement](https://software.intel.com/en-us/articles/end-user-license-agreement).
    The default value is `False`.

    icc: Boolean flag to specify whether the Intel C++ Compiler
    environment should be configured when `psxevars` is False.  This
    flag also controls whether to install the corresponding runtime in
    the `runtime` method.  Note: this flag does not control whether
    the developer environment is installed; see `components`.  The
    default is True.

    ifort: Boolean flag to specify whether the Intel Fortran Compiler
    environment should be configured when `psxevars` is False.  This
    flag also controls whether to install the corresponding runtime in
    the `runtime` method.  Note: this flag does not control whether
    the developer environment is installed; see `components`.  The
    default is True.

    ipp: Boolean flag to specify whether the Intel Integrated
    Performance Primitives environment should be configured when
    `psxevars` is False.  This flag also controls whether to install
    the corresponding runtime in the `runtime` method.  Note: this
    flag does not control whether the developer environment is
    installed; see `components`.  The default is True.

    license: The license to use to activate Intel Parallel Studio XE.
    If the string contains a `@` the license is interpreted as a
    network license, e.g., `12345@lic-server`.  Otherwise, the string
    is interpreted as the path to the license file relative to the
    local build context.  The default value is empty.  While this
    value is not required, the installation is unlikely to be
    successful without a valid license.

    mkl: Boolean flag to specify whether the Intel Math Kernel Library
    environment should be configured when `psxevars` is False.  This
    flag also controls whether to install the corresponding runtime in
    the `runtime` method.  Note: this flag does not control whether
    the developer environment is installed; see `components`.  The
    default is True.

    mpi: Boolean flag to specify whether the Intel MPI Library
    environment should be configured when `psxevars` is False.  This
    flag also controls whether to install the corresponding runtime in
    the `runtime` method.  Note: this flag does not control whether
    the developer environment is installed; see `components`.  The
    default is True.

    ospackages: List of OS packages to install prior to installing
    Intel MPI.  For Ubuntu, the default values are `build-essential`
    and `cpio`.  For RHEL-based Linux distributions, the default
    values are `gcc`, `gcc-c++`, `make`, and `which`.

    prefix: The top level install location.  The default value is
    `/opt/intel`.

    psxevars: Intel Parallel Studio XE provides an environment script
    (`compilervars.sh`) to setup the environment.  If this value is
    `True`, the bashrc is modified to automatically source this
    environment script.  However, the Intel runtime environment is not
    automatically available to subsequent container image build steps;
    the environment is available when the container image is run.  To
    set the Intel Parallel Studio XE environment in subsequent build
    steps you can explicitly call `source
    /opt/intel/compilers_and_libraries/linux/bin/compilervars.sh
    intel64` in each build step.  If this value is to set `False`,
    then the environment is set such that the environment is visible
    to both subsequent container image build steps and when the
    container image is run.  However, the environment may differ
    slightly from that set by `compilervars.sh`. This option will be
    used with the `runtime` method. The default value is
    `True`.

    runtime_version: The version of Intel Parallel Studio XE runtime
    to install via the `runtime` method.  The runtime is installed
    using the [intel_psxe_runtime](#intel_psxe_runtime) building
    block.  This value is passed as its `version` parameter.  In
    general, the major version of the runtime should correspond to the
    tarball version.  The default value is `2020.2-14`.

    tarball: Path to the Intel Parallel Studio XE tarball relative to
    the local build context.  The default value is empty.  This
    parameter is required.

    tbb: Boolean flag to specify whether the Intel Threading Building
    Blocks environment should be configured when `psxevars` is False.
    This flag also controls whether to install the corresponding
    runtime in the `runtime` method.  Note: this flag does not control
    whether the developer environment is installed; see `components`.
    The default is True.

    # Examples

    ```python
    intel_psxe(eula=True, license='XXXXXXXX.lic',
               tarball='parallel_studio_xe_2018_update1_professional_edition.tgz')
    ```

    ```python
    i = intel_psxe(...)
    openmpi(..., toolchain=i.toolchain, ...)
    ```

    """

    def __init__(self, **kwargs):
        """Initialize building block"""

        super(intel_psxe, self).__init__(**kwargs)

        # By setting this value to True, you agree to the
        # corresponding Intel End User License Agreement
        # (https://software.intel.com/en-us/articles/end-user-license-agreement)
        self.__eula = kwargs.get('eula', False)

        self.__components = kwargs.get('components', ['DEFAULTS'])
        self.__daal = kwargs.get('daal', True)
        self.__icc = kwargs.get('icc', True)
        self.__ifort = kwargs.get('ifort', True)
        self.__ipp = kwargs.get('ipp', True)
        self.__license = kwargs.get('license', None)
        self.__mkl = kwargs.get('mkl', True)
        self.__mpi = kwargs.get('mpi', True)
        self.__ospackages = kwargs.get('ospackages', [])
        self.__prefix = kwargs.get('prefix', '/opt/intel')
        self.__psxevars = kwargs.get('psxevars', True)
        self.__runtime_version = kwargs.get('runtime_version', '2020.2-14')
        self.__tarball = kwargs.get('tarball', None)
        self.__tbb = kwargs.get('tbb', True)
        self.__wd = kwargs.get('wd', hpccm.config.g_wd) # working directory

        self.toolchain = toolchain(CC='icc', CXX='icpc', F77='ifort',
                                   F90='ifort', FC='ifort')
        self.toolchain.CFLAGS = hpccm.config.get_cpu_optimization_flags('intel')
        self.toolchain.CXXFLAGS = hpccm.config.get_cpu_optimization_flags('intel')
        self.toolchain.FFLAGS = hpccm.config.get_cpu_optimization_flags('intel')
        self.toolchain.FCFLAGS = hpccm.config.get_cpu_optimization_flags('intel')

        self.__bashrc = ''   # Filled in by __distro()
        self.__commands = [] # Filled in by __setup()

        if hpccm.config.g_cpu_arch != cpu_arch.X86_64: # pragma: no cover
            logging.warning('Using intel_psxe on a non-x86_64 processor')

        # Set the Linux distribution specific parameters
        self.__distro()

        # Construct the series of steps to execute
        self.__setup()

        # Fill in container instructions
        self.__instructions()

    def __instructions(self):
        """Fill in container instructions"""

        self += comment('Intel Parallel Studio XE')
        self += packages(ospackages=self.__ospackages)
        self += copy(src=self.__tarball,
                     dest=posixpath.join(self.__wd, self.__tarball_name))
        if self.__license and not '@' in self.__license:
            # License file
            self += copy(src=self.__license,
                         dest=posixpath.join(self.__wd, 'license.lic'))
        self += shell(commands=self.__commands)

        if self.__psxevars:
            # Source the mpivars environment script when starting the
            # container, but the variables not be available for any
            # subsequent build steps.
            self += shell(commands=['echo "source {0}/compilers_and_libraries/linux/bin/compilervars.sh intel64" >> {1}'.format(self.__prefix, self.__bashrc)])
        else:
            self += environment(variables=self.environment_step())

    def __distro(self):
        """Based on the Linux distribution, set values accordingly.  A user
        specified value overrides any defaults."""

        if hpccm.config.g_linux_distro == linux_distro.UBUNTU:
            if not self.__ospackages:
                self.__ospackages = ['build-essential', 'cpio']
            self.__bashrc = '/etc/bash.bashrc'
        elif hpccm.config.g_linux_distro == linux_distro.CENTOS:
            if not self.__ospackages:
                self.__ospackages = ['gcc', 'gcc-c++', 'make', 'which']
            self.__bashrc = '/etc/bashrc'
        else: # pragma: no cover
            raise RuntimeError('Unknown Linux distribution')

    def __environment(self):
        basepath = posixpath.join(self.__prefix, 'compilers_and_libraries',
                                  'linux')
        cpath = []
        ld_library_path = []
        library_path = []
        path = []
        env = {}

        if self.__daal:
            env['DAALROOT'] = posixpath.join(basepath, 'daal')
            cpath.append(posixpath.join(basepath, 'daal', 'include'))
            ld_library_path.append(posixpath.join(basepath, 'daal', 'lib',
                                                  'intel64'))
            library_path.append(posixpath.join(basepath, 'daal', 'lib',
                                               'intel64'))

        if self.__icc:
            cpath.append(posixpath.join(basepath, 'pstl', 'include'))
            ld_library_path.append(posixpath.join(basepath, 'compiler', 'lib',
                                                  'intel64'))
            path.append(posixpath.join(basepath, 'bin', 'intel64'))

        if self.__ifort:
            ld_library_path.append(posixpath.join(basepath, 'compiler', 'lib',
                                                  'intel64'))
            path.append(posixpath.join(basepath, 'bin', 'intel64'))

        if self.__ipp:
            env['IPPROOT' ] = posixpath.join(basepath, 'ipp')
            cpath.append(posixpath.join(basepath, 'ipp', 'include'))
            ld_library_path.append(posixpath.join(basepath, 'ipp', 'lib',
                                                  'intel64'))
            library_path.append(posixpath.join(basepath, 'ipp', 'lib',
                                               'intel64'))

        if self.__mkl:
            env['MKLROOT'] = posixpath.join(basepath, 'mkl')
            cpath.append(posixpath.join(basepath, 'mkl', 'include'))
            ld_library_path.append(posixpath.join(basepath, 'mkl', 'lib',
                                                'intel64'))
            library_path.append(posixpath.join(basepath, 'mkl', 'lib',
                                               'intel64'))

        if self.__mpi:
            # Handle libfabics case
            env['I_MPI_ROOT' ] = posixpath.join(basepath, 'mpi')
            cpath.append(posixpath.join(basepath, 'mpi', 'include'))
            ld_library_path.append(posixpath.join(basepath, 'mpi', 'intel64',
                                                  'lib'))
            path.append(posixpath.join(basepath, 'mpi', 'intel64', 'bin'))

        if self.__tbb:
            cpath.append(posixpath.join(basepath, 'tbb', 'include'))
            ld_library_path.append(posixpath.join(basepath, 'tbb', 'lib',
                                                  'intel64', 'gcc4.7'))
            library_path.append(posixpath.join(basepath, 'tbb', 'lib',
                                               'intel64', 'gcc4.7'))

        if cpath:
            cpath.append('$CPATH')
            env['CPATH'] = ':'.join(cpath)

        if library_path:
            library_path.append('$LIBRARY_PATH')
            env['LIBRARY_PATH'] = ':'.join(library_path)

        if ld_library_path:
            ld_library_path.append('$LD_LIBRARY_PATH')
            env['LD_LIBRARY_PATH'] = ':'.join(ld_library_path)

        if path:
            path.append('$PATH')
            env['PATH'] = ':'.join(path)

        return env

    def __setup(self):
        """Construct the series of shell commands, i.e., fill in
           self.__commands"""

        # tarball must be specified
        if not self.__tarball:
            raise RuntimeError('Intel PSXE tarball not specified')

        # Get the name of the directory that created when the tarball
        # is extracted.  Assume it is the same as the basename of the
        # tarball.
        self.__tarball_name = posixpath.basename(self.__tarball)
        basedir = posixpath.splitext(self.__tarball_name)[0]

        # Untar
        self.__commands.append(self.untar_step(
            tarball=posixpath.join(self.__wd, self.__tarball_name),
            directory=(self.__wd)))

        # Configure silent install
        silent_cfg=[
            r's/^#\?\(COMPONENTS\)=.*/\1={}/g'.format(
                ';'.join(self.__components)),
            r's|^#\?\(PSET_INSTALL_DIR\)=.*|\1={}|g'.format(self.__prefix)]

        # EULA acceptance
        if self.__eula:
            silent_cfg.append(r's/^#\?\(ACCEPT_EULA\)=.*/\1=accept/g')

        # License activation
        if self.__license and '@' in self.__license:
            # License server
            silent_cfg.append(r's/^#\?\(ACTIVATION_TYPE\)=.*/\1=license_server/g')
            silent_cfg.append(r's/^#\?\(ACTIVATION_LICENSE_FILE\)=.*/\1={}/g'.format(self.__license))
        elif self.__license:
            # License file
            silent_cfg.append(r's/^#\?\(ACTIVATION_TYPE\)=.*/\1=license_file/g')
            silent_cfg.append(r's|^#\?\(ACTIVATION_LICENSE_FILE\)=.*|\1={}|g'.format(posixpath.join(self.__wd, 'license.lic')))
        else:
            # No license, will most likely not work
            logging.warning('No Intel Parallel Studio XE license specified')

        # Update the silent config file
        self.__commands.append(self.sed_step(
            file=posixpath.join(self.__wd, basedir, 'silent.cfg'),
            patterns=silent_cfg))

        # Install
        self.__commands.append(
            'cd {} && ./install.sh --silent=silent.cfg'.format(
                posixpath.join(self.__wd, basedir)))

        # Cleanup runfile
        self.__commands.append(self.cleanup_step(
            items=[posixpath.join(self.__wd, self.__tarball_name),
                   posixpath.join(self.__wd, basedir)]))

        # Set the environment
        self.environment_variables = self.__environment()

    def runtime(self, _from='0'):
        """Install the runtime from a full build in a previous stage"""
        return str(intel_psxe_runtime(daal=self.__daal,
                                      eula=self.__eula,
                                      icc=self.__icc,
                                      ifort=self.__ifort,
                                      ipp=self.__ipp,
                                      mkl=self.__mkl,
                                      mpi=self.__mpi,
                                      psxevars=self.__psxevars,
                                      tbb=self.__tbb,
                                      version=self.__runtime_version))



# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, too-few-public-methods
# pylint: disable=too-many-instance-attributes

"""NVSHMEM building block"""

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

import os
import posixpath

import hpccm.templates.downloader
import hpccm.templates.envvars
import hpccm.templates.rm
import hpccm.templates.tar

from hpccm.building_blocks.base import bb_base
from hpccm.building_blocks.generic_build import generic_build
from hpccm.building_blocks.packages import packages
from hpccm.primitives.comment import comment
from hpccm.primitives.copy import copy
from hpccm.primitives.environment import environment
from hpccm.primitives.shell import shell

class nvshmem(bb_base, hpccm.templates.downloader, hpccm.templates.envvars,
              hpccm.templates.rm, hpccm.templates.tar):
    """The `nvshmem` building block builds and installs the
    [NVSHMEM](https://developer.nvidia.com/nvshmem) component.

    # Parameters

    binary_tarball: Path to NVSHMEM binary tarball relative to the
    build context. The default value is empty. Either this parameter
    or `package` must be specified.

    cuda: Flag to specify the path to the CUDA installation.  The
    default is `/usr/local/cuda`.

    environment: Boolean flag to specify whether the environment
    (`CPATH`, `LIBRARY_PATH`, and `PATH`) should be modified to
    include NVSHMEM. The default is True.

    gdrcopy: Flag to specify the path to the GDRCOPY installation.
    The default is empty.

    hydra: Boolean flag to specify whether the Hydra process launcher
    should be installed.  If True, adds `automake` to the list of OS
    packages.  The default is False.

    make_variables: Dictionary of environment variables and values to
    set when building NVSHMEM.  The default is an empty dictionary.

    mpi: Flag to specify the path to the MPI installation.  The
    default is empty, i.e., do not build NVSHMEM with MPI support.

    ospackages: List of OS packages to install prior to building.  The
    default values are `make` and `wget`.

    package: Path to the NVSHMEM source package relative to the build
    context. The default value is empty. Either this parameter or
    `binary_tarball` must be specified.

    prefix: The top level install location.  The default value is
    `/usr/local/nvshmem`.

    shmem: Flag to specify the path to the SHMEM installation.  The
    default is empty, i.e., do not build NVSHMEM with SHMEM support.

    version: The version of NVSHMEM source to download.  The default
    value is `2.1.2`.

    # Examples

    ```python
    nvshmem(mpi='/usr/local/openmpi', version='2.1.2')
    ```

    """

    def __init__(self, **kwargs):
        """Initialize building block"""

        super(nvshmem, self).__init__(**kwargs)

        self.__binary_tarball = kwargs.pop('binary_tarball', None)
        self.__cuda = kwargs.pop('cuda', '/usr/local/cuda')
        self.__gdrcopy = kwargs.pop('gdrcopy', None)
        self.__hydra = kwargs.pop('hydra', False)
        self.__make_variables = kwargs.pop('make_variables', {})
        self.__mpi = kwargs.pop('mpi', None)
        self.__ospackages = kwargs.pop('ospackages', ['make', 'wget'])
        self.__prefix = kwargs.pop('prefix', '/usr/local/nvshmem')
        self.__release = kwargs.pop('release', '0')
        self.__shmem = kwargs.pop('shmem', None)
        self.__src_directory = kwargs.pop('src_directory', None)
        self.__version = kwargs.pop('version', '2.1.2')
        self.__wd = kwargs.get('wd', hpccm.config.g_wd) # working directory

        # Set the download specific parameters
        self.__download()
        kwargs['url'] = self.url
        if self.__src_directory:
            kwargs['directory'] = self.__src_directory

        # Setup the environment variables
        self.environment_variables['CPATH'] = '{}:$CPATH'.format(
            posixpath.join(self.__prefix, 'include'))
        self.environment_variables['LIBRARY_PATH'] = '{}:$LIBRARY_PATH'.format(
            posixpath.join(self.__prefix, 'lib'))
        self.environment_variables['PATH'] = '{}:$PATH'.format(
            posixpath.join(self.__prefix, 'bin'))

        # Add packages
        if self.__hydra:
            self.__ospackages.append('automake')

        if self.__version and not self.__binary_tarball and not self.package:
            self += comment('NVSHMEM {}'.format(self.__version))
        else:
            self += comment('NVSHMEM')
        self += packages(ospackages=self.__ospackages)

        if self.__binary_tarball:
            # Shorthand for the tarball file inside the container
            tarball = posixpath.join(self.__wd,
                                     os.path.basename(self.__binary_tarball))

            self += copy(src=self.__binary_tarball, dest=tarball)
            self += shell(commands=[
                # Untar binary package
                self.untar_step(
                    tarball=tarball,
                    # remove the leading directory, e.g., install in
                    # /usr/local/nvshmem not
                    # /usr/local/nvshmem/nvshmem_<version>_<arch>.
                    args=['--strip-components=1'],
                    directory=self.__prefix),
                # Install Hydra process launcher
                '{0}/scripts/install_hydra.sh {1} {0}'.format(
                    self.__prefix, self.__wd) if self.__hydra else None,
                # Remove temporary files and cleanup
                self.cleanup_step(items=[tarball])])
            self += environment(variables=self.environment_variables)

        else:
            # Build from source

            # Set the build options
            self.__configure()

            self.__bb = generic_build(
                build = [
                    '{} make -j$(nproc) install'.format(
                        self.__build_environment),
                    './scripts/install_hydra.sh {1} {0}'.format(
                        self.__prefix, self.__wd) if self.__hydra else None],
                comment=False,
                devel_environment=self.environment_variables,
                prefix=self.__prefix,
                runtime_environment=self.environment_variables,
                **kwargs)
            self += self.__bb

    def __configure(self):
        """Setup build options based on user parameters"""

        e = {}

        e['NVSHMEM_PREFIX'] = self.__prefix

        # Default to 0 unless MPI/SHMEM is requested
        e['NVSHMEM_MPI_SUPPORT'] = 0

        if self.__cuda:
            e['CUDA_HOME'] = self.__cuda

        if self.__gdrcopy:
            e['GDRCOPY_HOME'] = self.__gdrcopy

        if self.__mpi:
            e['NVSHMEM_MPI_SUPPORT'] = 1
            e['MPI_HOME'] = self.__mpi

        if self.__shmem:
            e['NVSHMEM_SHMEM_SUPPORT'] = 1
            e['SHMEM_HOME'] = self.__shmem

        if self.__make_variables:
          e.update(self.__make_variables)

        l = []
        if e:
            for key, val in sorted(e.items()):
                l.append('{0}={1}'.format(key, val))

        self.__build_environment = ' '.join(l)

    def __download(self):
        """Set download source based on user parameters"""

        if not self.package and not self.repository and not self.url:
            self.url = 'https://developer.download.nvidia.com/compute/redist/nvshmem/{0}/source/nvshmem_src_{0}-{1}.txz'.format(self.__version, self.__release)

    def runtime(self, _from='0'):
        """Generate the set of instructions to install the runtime specific
        components from a build in a previous stage.

        # Examples

        ```python
        n = nvshmem(...)
        Stage0 += n
        Stage1 += n.runtime()
        ```
        """
        self.rt += comment('NVSHMEM')
        if self.__binary_tarball:
            self.rt += copy(_from=_from, src=self.__prefix, dest=self.__prefix)
            self.rt += environment(variables=self.environment_step())
        else:
            self.rt += self.__bb.runtime(_from=_from)
        return str(self.rt)
